import json
import sys
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

import aiohttp
import requests
import yarl

from fused._deserialize_parquet import parquet_to_df
from fused._optional_deps import (
    GPD_GEODATAFRAME,
    HAS_GEOPANDAS,
    HAS_PANDAS,
    PD_DATAFRAME,
)
from fused._options import options as OPTIONS
from fused.models.udf._eval_result import UdfEvaluationResult
from fused.types import UdfRuntimeError

from ._impl._context_impl import context_get_auth_header, context_get_user_email
from ._impl._realtime_ops_impl import (
    get_recursion_factor,
    make_realtime_url,
    make_shared_realtime_url,
)
from ._serialization import deserialize_png, deserialize_tiff

if TYPE_CHECKING:
    import pandas as pd
    import xarray as xr

DEFAULT_DTYPE_VECTOR = "parquet"
DEFAULT_DTYPE_RASTER = "tiff"


def serialize_realtime_params(params: Dict[str, Any]):
    result = {}
    for param_name, param_value in params.items():
        if (HAS_GEOPANDAS and isinstance(param_value, GPD_GEODATAFRAME)) or (
            HAS_PANDAS and isinstance(param_value, PD_DATAFRAME)
        ):
            result[param_name] = param_value.to_json()
        elif isinstance(param_value, (list, tuple, dict, bool)):
            result[param_name] = json.dumps(param_value)
        else:
            result[param_name] = param_value

    return result


def _parse_realtime_response(r: requests.Response, content_type: str):
    if content_type == "application/octet-stream":  # parquet
        data = parquet_to_df(r.content)
    elif content_type == "image/png":
        data = deserialize_png(r.content)
    elif content_type == "image/tiff":
        data = deserialize_tiff(r.content)
    else:
        data = r.content  # TODO

    return UdfEvaluationResult(
        data=data,
        time_taken_seconds=0.0,  # TODO
    )


async def _parse_realtime_response_async(content: bytes, content_type: str):
    if content_type == "application/octet-stream":  # parquet
        data = parquet_to_df(content)
    elif content_type == "image/png":
        data = deserialize_png(content)
    elif content_type == "image/tiff":
        data = deserialize_tiff(content)
    else:
        data = content  # TODO

    return data


def _realtime_raise_for_status(r: requests.Response):
    if r.status_code >= 400 and "x-fused-error" in r.headers:
        msg = str(r.headers["x-fused-error"])
        raise requests.HTTPError(msg, response=r)
    if r.status_code >= 400 and "x-fused-metadata" in r.headers:
        # x-fused-metadata holds LogHandler as JSON, which contains stdout/stderr.
        _fused_metadata = r.headers.get("x-fused-metadata")
        fused_metadata = json.loads(_fused_metadata) if _fused_metadata else {}

        # Extract stdout/stderr.
        stdout = fused_metadata.get("stdout")
        stderr = fused_metadata.get("stderr")

        # Extract exception details
        exception_class = fused_metadata.get("exception_class")
        has_exception = fused_metadata.get("has_exception", False)

        # Extract error line, if exists.
        error_lineno = fused_metadata.get("lineno")

        error_message = ""
        if "errormsg" in fused_metadata and fused_metadata["errormsg"]:
            error_message = f"The UDF returned the following error in line {fused_metadata.get('lineno')}:\n{fused_metadata['errormsg']}"
        elif "exception" in fused_metadata and fused_metadata["exception"]:
            error_message = fused_metadata["exception"]

        return UdfEvaluationResult(
            data=None,
            time_taken_seconds=0.0,  # TODO
            stdout=stdout,
            stderr=stderr,
            error_message=error_message,
            error_lineno=error_lineno,
            has_exception=has_exception,
            exception_class=exception_class,
        )

    r.raise_for_status()


def _realtime_raise_for_status_async(r: aiohttp.ClientResponse):
    if r.status >= 400 and "x-fused-error" in r.headers:
        msg = str(r.headers["x-fused-error"])
        raise aiohttp.ClientResponseError(
            request_info=None,
            status=r.status,
            message=msg,
            headers=[],
            history=(),
        )
    if r.status >= 400 and "x-fused-metadata" in r.headers:
        # x-fused-metadata holds LogHandler as JSON, which contains stdout/stderr.
        _fused_metadata = r.headers.get("x-fused-metadata")
        fused_metadata = json.loads(_fused_metadata) if _fused_metadata else {}

        # Extract stdout/stderr.
        stdout = fused_metadata.get("stdout")
        # stderr = fused_metadata.get("stderr")

        # Extract exception details
        exception_class = fused_metadata.get("exception_class")
        # has_exception = fused_metadata.get("has_exception", False)

        # Extract error line, if exists.
        # error_lineno = fused_metadata.get("lineno")

        error_message = ""
        if "errormsg" in fused_metadata and fused_metadata["errormsg"]:
            error_message = f"The UDF returned the following error in line {fused_metadata.get('lineno')}:\n{fused_metadata['errormsg']}"
        elif "exception" in fused_metadata and fused_metadata["exception"]:
            error_message = fused_metadata["exception"]

        if stdout:
            sys.stdout.write(stdout)
        raise UdfRuntimeError(error_message, child_exception_class=exception_class)

    r.raise_for_status()


async def _realtime_follow_redirect_async(
    *, session: aiohttp.ClientSession, r: aiohttp.ClientResponse
):
    if r.status >= 300 and r.status < 400 and "location" in r.headers:
        # Per this link, aiohttp will mangle the redirect URL
        # https://stackoverflow.com/questions/77319421/aiohttp-showing-403-forbidden-error-but-requests-get-giving-200-ok-response
        url = yarl.URL(r.headers["location"], encoded=True)
        return await session.get(url)
    return r


def _run_tile(
    url: str,
    *,
    cache_max_age: Optional[int] = None,
    _dtype_out_vector: str = DEFAULT_DTYPE_VECTOR,
    _dtype_out_raster: str = DEFAULT_DTYPE_RASTER,
    _is_shared: bool = False,
    **params,
):
    access_token_header = context_get_auth_header(missing_ok=_is_shared)
    recursion_factor = get_recursion_factor()

    req_params = {
        # TODO...
        **({"cache_max_age": int(cache_max_age)} if cache_max_age is not None else {}),
        "dtype_out_vector": _dtype_out_vector,
        "dtype_out_raster": _dtype_out_raster,
        **(serialize_realtime_params(params) if params is not None else {}),
    }

    r = requests.get(
        url=url,
        params=req_params,
        headers={
            **access_token_header,
            "Fused-Recursion": f"{recursion_factor}",
        },
    )
    exc = _realtime_raise_for_status(r)
    if exc is not None:
        return exc

    return _parse_realtime_response(r, content_type=r.headers["content-type"])


def run_tile(
    email: str,
    id: Optional[str] = None,
    *,
    x: int,
    y: int,
    z: int,
    cache_max_age: Optional[int] = None,
    _dtype_out_vector: str = DEFAULT_DTYPE_VECTOR,
    _dtype_out_raster: str = DEFAULT_DTYPE_RASTER,
    _client_id: Optional[str] = None,
    **params,
) -> Optional[Union["pd.DataFrame", "xr.Dataset"]]:
    """
    Executes a private tile-based UDF indexed under the specified email and ID. The calling user must have the necessary permissions to execute the UDF.

    This function constructs a URL to run a UDF on a specific tile defined by its x, y, and z coordinates, and
    sends a request to the server. It supports customization of the output data types for vector and raster data,
    as well as additional parameters for the UDF execution.

    Args:
        email (str): Email address of user account associated with the UDF.
        id (Optional[str]): Unique identifier for the UDF. If None, the user's email is used as the ID.
        x (int): The x coordinate of the tile.
        y (int): The y coordinate of the tile.
        z (int): The zoom level of the tile.
        cache_max_age (Optional[int]): The maximum age when returning a result from the cache.
        _dtype_out_vector (str): Desired data type for vector output. Defaults to a pre-defined type.
        _dtype_out_raster (str): Desired data type for raster output. Defaults to a pre-defined type.
        _client_id (Optional[str]): Client identifier for API usage. If None, a default or global client ID may be used.
        **params: Additional keyword arguments for the UDF execution.

    Returns:
        The response from the server after executing the UDF on the specified tile.
    """
    if id is None:
        id = email
        email = context_get_user_email()

    url = f"{make_realtime_url(_client_id)}/api/v1/run/udf/saved/{email}/{id}/tiles/{z}/{x}/{y}"
    return _run_tile(
        url=url,
        cache_max_age=cache_max_age,
        _dtype_out_vector=_dtype_out_vector,
        _dtype_out_raster=_dtype_out_raster,
        **params,
    )


def run_shared_tile(
    token: str,
    *,
    x: int,
    y: int,
    z: int,
    cache_max_age: Optional[int] = None,
    _dtype_out_vector: str = DEFAULT_DTYPE_VECTOR,
    _dtype_out_raster: str = DEFAULT_DTYPE_RASTER,
    _client_id: Optional[str] = None,
    **params,
) -> Optional[Union["pd.DataFrame", "xr.Dataset"]]:
    """
    Executes a shared tile-based UDF.

    This function constructs a URL to run a UDF on a specific tile defined by its x, y, and z coordinates, and
    sends a request to the server. It supports customization of the output data types for vector and raster data,
    as well as additional parameters for the UDF execution.

    Args:
        token (str): A shared access token that authorizes the operation.
        id (Optional[str]): Unique identifier for the UDF. If None, the user's email is used as the ID.
        x (int): The x coordinate of the tile.
        y (int): The y coordinate of the tile.
        z (int): The zoom level of the tile.
        cache_max_age (Optional[int]): The maximum age when returning a result from the cache.
        _dtype_out_vector (str): Desired data type for vector output. Defaults to a pre-defined type.
        _dtype_out_raster (str): Desired data type for raster output. Defaults to a pre-defined type.
        _client_id (Optional[str]): Client identifier for API usage. If None, a default or global client ID may be used.
        **params: Additional keyword arguments for the UDF execution.

    Returns:
        The response from the server after executing the UDF on the specified tile.
    """
    url = f"{make_shared_realtime_url(token)}/run/tiles/{z}/{x}/{y}"
    return _run_tile(
        url=url,
        cache_max_age=cache_max_age,
        _dtype_out_vector=_dtype_out_vector,
        _dtype_out_raster=_dtype_out_raster,
        _client_id=_client_id,
        _is_shared=True,
        **params,
    )


def _run_file(
    url: str,
    *,
    cache_max_age: Optional[int] = None,
    _dtype_out_vector: str = DEFAULT_DTYPE_VECTOR,
    _dtype_out_raster: str = DEFAULT_DTYPE_RASTER,
    _is_shared: bool = False,
    **params,
):
    access_token_header = context_get_auth_header(missing_ok=_is_shared)
    recursion_factor = get_recursion_factor()

    req_params = {
        # TODO...
        **({"cache_max_age": int(cache_max_age)} if cache_max_age is not None else {}),
        "dtype_out_vector": _dtype_out_vector,
        "dtype_out_raster": _dtype_out_raster,
        **(serialize_realtime_params(params) if params is not None else {}),
    }

    r = requests.get(
        url=url,
        params=req_params,
        headers={
            **access_token_header,
            "Fused-Recursion": f"{recursion_factor}",
        },
    )
    exc = _realtime_raise_for_status(r)
    if exc is not None:
        return exc

    return _parse_realtime_response(r, content_type=r.headers["content-type"])


def run_file(
    email: str,
    id: Optional[str] = None,
    *,
    cache_max_age: Optional[int] = None,
    _dtype_out_vector: str = DEFAULT_DTYPE_VECTOR,
    _dtype_out_raster: str = DEFAULT_DTYPE_RASTER,
    _client_id: Optional[str] = None,
    **params,
) -> Optional[Union["pd.DataFrame", "xr.Dataset"]]:
    """
    Executes a private file-based UDF indexed under the specified email and ID. The calling user must have the necessary permissions to execute the UDF.

    This function constructs a URL to run a UDF associated with the given email and ID, allowing for output data type customization for both vector and raster outputs. It also supports additional parameters for the UDF execution.

    Args:
        email (str): Email address of user account associated with the UDF.
        id (Optional[str]): Unique identifier for the UDF. If None, the user's email is used as the ID.
        cache_max_age (Optional[int]): The maximum age when returning a result from the cache.
        _dtype_out_vector (str): Desired data type for vector output, defaults to a predefined type.
        _dtype_out_raster (str): Desired data type for raster output, defaults to a predefined type.
        _client_id (Optional[str]): Client identifier for API usage. If None, a default or global client ID may be used.
        **params: Additional keyword arguments for the UDF execution.

    Returns:
        The response from the server after executing the UDF.
    """
    if id is None:
        id = email
        email = context_get_user_email()

    url = f"{make_realtime_url(_client_id)}/api/v1/run/udf/saved/{email}/{id}"
    return _run_file(
        url=url,
        cache_max_age=cache_max_age,
        _dtype_out_vector=_dtype_out_vector,
        _dtype_out_raster=_dtype_out_raster,
        **params,
    )


def run_shared_file(
    token: str,
    *,
    cache_max_age: Optional[int] = None,
    _dtype_out_vector: str = DEFAULT_DTYPE_VECTOR,
    _dtype_out_raster: str = DEFAULT_DTYPE_RASTER,
    **params,
) -> Optional[Union["pd.DataFrame", "xr.Dataset"]]:
    """
    Executes a shared file-based UDF.

    This function constructs a URL for running an operation on a file accessible via a shared token. It allows for customization of the output data types for vector and raster data and supports additional parameters for the operation's execution.

    Args:
        token (str): A shared access token that authorizes the operation.
        cache_max_age (Optional[int]): The maximum age when returning a result from the cache.
        _dtype_out_vector (str): Desired data type for vector output, defaults to a predefined type.
        _dtype_out_raster (str): Desired data type for raster output, defaults to a predefined type.
        **params: Additional keyword arguments for the operation execution.

    Returns:
        The response from the server after executing the operation on the file.

    Raises:
        Exception: Describes various exceptions that could occur during the function execution, including but not limited to invalid parameters, network errors, unauthorized access errors, or server-side errors.

    Note:
        This function is designed to access shared operations that require a token for authorization. It requires network access to communicate with the server hosting these operations and may incur data transmission costs or delays depending on the network's performance.
    """
    url = f"{make_shared_realtime_url(token)}/run/file"
    return _run_file(
        url=url,
        cache_max_age=cache_max_age,
        _dtype_out_vector=_dtype_out_vector,
        _dtype_out_raster=_dtype_out_raster,
        _is_shared=True,
        **params,
    )


async def _run_tile_async(
    url: str,
    *,
    cache_max_age: Optional[int] = None,
    _dtype_out_vector: str = DEFAULT_DTYPE_VECTOR,
    _dtype_out_raster: str = DEFAULT_DTYPE_RASTER,
    _is_shared: bool = False,
    **params,
):
    access_token_header = context_get_auth_header(missing_ok=_is_shared)
    recursion_factor = get_recursion_factor()

    req_params = {
        # TODO...
        **({"cache_max_age": int(cache_max_age)} if cache_max_age is not None else {}),
        "dtype_out_vector": _dtype_out_vector,
        "dtype_out_raster": _dtype_out_raster,
        **(serialize_realtime_params(params) if params is not None else {}),
    }

    if OPTIONS.pyodide_async_requests:
        import pyodide.http

        url_with_params = yarl.URL(url, encoded=True).with_query(req_params)
        r = await pyodide.http.pyfetch(
            str(url_with_params),
            headers={
                **access_token_header,
                "Fused-Recursion": f"{recursion_factor}",
            },
        )
        redirect = r.headers.get("x-fused-redirect", None)
        if redirect:
            url = yarl.URL(redirect, encoded=True)
            r = await pyodide.http.pyfetch(url)

        exc = _realtime_raise_for_status_async(r)
        if exc is not None:
            return exc

        return await _parse_realtime_response_async(
            await r.bytes(), content_type=r.headers["content-type"]
        )
    else:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                url=url,
                params=req_params,
                headers={
                    **access_token_header,
                    "Fused-Recursion": f"{recursion_factor}",
                },
                allow_redirects=False,
            ) as r:
                r = await _realtime_follow_redirect_async(session=session, r=r)
                exc = _realtime_raise_for_status_async(r)
                if exc is not None:
                    return exc

                return await _parse_realtime_response_async(
                    await r.read(), content_type=r.headers["content-type"]
                )


async def run_tile_async(
    email: str,
    id: Optional[str] = None,
    *,
    x: int,
    y: int,
    z: int,
    cache_max_age: Optional[int] = None,
    _dtype_out_vector: str = DEFAULT_DTYPE_VECTOR,
    _dtype_out_raster: str = DEFAULT_DTYPE_RASTER,
    _client_id: Optional[str] = None,
    **params,
) -> Optional[Union["pd.DataFrame", "xr.Dataset"]]:
    """
    Asynchronously executes a private tile-based UDF indexed under the specified email and ID. The calling user must have the necessary permissions to execute the UDF.

    This function constructs a URL to asynchronously run a UDF on a specific tile defined by its x, y, and z coordinates. It supports customization of the output data types for vector and raster data, and accommodates additional parameters for the UDF execution.

    Args:
        email (str): User's email address. Used to identify the user's saved UDFs. If the ID is not provided, the email is also used as the ID.
        id (Optional[str]): Unique identifier for the UDF. If None, the user's email is used as the ID.
        x (int): The x coordinate of the tile.
        y (int): The y coordinate of the tile.
        z (int): The zoom level of the tile.
        cache_max_age (Optional[int]): The maximum age when returning a result from the cache.
        _dtype_out_vector (str): Desired data type for vector output. Defaults to a predefined type.
        _dtype_out_raster (str): Desired data type for raster output. Defaults to a predefined type.
        _client_id (Optional[str]): Client identifier for API usage. If None, a default or global client ID may be used.
        **params: Additional keyword arguments for the UDF execution.

    Returns:
        A coroutine that, when awaited, sends a request to the server to execute the UDF on the specified tile and returns the server's response. The format and content of the response depend on the UDF's implementation and the server's response format.
    """
    if id is None:
        id = email
        email = context_get_user_email()

    url = f"{make_realtime_url(_client_id)}/api/v1/run/udf/saved/{email}/{id}/tiles/{z}/{x}/{y}"
    return await _run_tile_async(
        url=url,
        cache_max_age=cache_max_age,
        _dtype_out_vector=_dtype_out_vector,
        _dtype_out_raster=_dtype_out_raster,
        **params,
    )


async def run_shared_tile_async(
    token: str,
    *,
    x: int,
    y: int,
    z: int,
    cache_max_age: Optional[int] = None,
    _dtype_out_vector: str = DEFAULT_DTYPE_VECTOR,
    _dtype_out_raster: str = DEFAULT_DTYPE_RASTER,
    **params,
) -> Optional[Union["pd.DataFrame", "xr.Dataset"]]:
    """
    Asynchronously executes a shared tile-based UDF using a specific access token.

    This function constructs a URL for running an operation on a tile, defined by its x, y, and z coordinates, accessible via a shared token. It allows for customization of the output data types for vector and raster data and supports additional parameters for the operation's execution.

    Args:
        token (str): A shared access token that authorizes the operation on the specified tile.
        x (int): The x coordinate of the tile.
        y (int): The y coordinate of the tile.
        z (int): The zoom level of the tile.
        cache_max_age (Optional[int]): The maximum age when returning a result from the cache.
        _dtype_out_vector (str): Desired data type for vector output, defaults to a predefined type.
        _dtype_out_raster (str): Desired data type for raster output, defaults to a predefined type.
        **params: Additional keyword arguments for the operation execution.

    Returns:
        A coroutine that, when awaited, sends a request to the server to execute the operation on the specified tile and returns the server's response. The format and content of the response depend on the operation's implementation and the server's response format.
    """
    url = f"{make_shared_realtime_url(token)}/run/tiles/{z}/{x}/{y}"
    return await _run_tile_async(
        url=url,
        cache_max_age=cache_max_age,
        _dtype_out_vector=_dtype_out_vector,
        _dtype_out_raster=_dtype_out_raster,
        _is_shared=True,
        **params,
    )


async def _run_file_async(
    url: str,
    *,
    cache_max_age: Optional[int] = None,
    _dtype_out_vector: str = DEFAULT_DTYPE_VECTOR,
    _dtype_out_raster: str = DEFAULT_DTYPE_RASTER,
    _is_shared: bool = False,
    **params,
):
    access_token_header = context_get_auth_header(missing_ok=_is_shared)
    recursion_factor = get_recursion_factor()

    req_params = {
        # TODO...
        **({"cache_max_age": int(cache_max_age)} if cache_max_age is not None else {}),
        "dtype_out_vector": _dtype_out_vector,
        "dtype_out_raster": _dtype_out_raster,
        **(serialize_realtime_params(params) if params is not None else {}),
    }

    if OPTIONS.pyodide_async_requests:
        import pyodide.http

        url_with_params = yarl.URL(url, encoded=True).with_query(req_params)
        r = await pyodide.http.pyfetch(
            str(url_with_params),
            headers={
                **access_token_header,
                "Fused-Recursion": f"{recursion_factor}",
            },
        )
        redirect = r.headers.get("x-fused-redirect", None)
        if redirect:
            url = yarl.URL(redirect, encoded=True)
            r = await pyodide.http.pyfetch(url)

        exc = _realtime_raise_for_status_async(r)
        if exc is not None:
            return exc

        return await _parse_realtime_response_async(
            await r.bytes(), content_type=r.headers["content-type"]
        )
    else:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                url=url,
                params=req_params,
                headers={
                    **access_token_header,
                    "Fused-Recursion": f"{recursion_factor}",
                },
                allow_redirects=False,
            ) as r:
                r = await _realtime_follow_redirect_async(session=session, r=r)
                exc = _realtime_raise_for_status_async(r)
                if exc is not None:
                    return exc

                return await _parse_realtime_response_async(
                    await r.read(), content_type=r.headers["content-type"]
                )


async def run_file_async(
    email: str,
    id: Optional[str] = None,
    *,
    cache_max_age: Optional[int] = None,
    _dtype_out_vector: str = DEFAULT_DTYPE_VECTOR,
    _dtype_out_raster: str = DEFAULT_DTYPE_RASTER,
    _client_id: Optional[str] = None,
    **params,
) -> Optional[Union["pd.DataFrame", "xr.Dataset"]]:
    """
    Asynchronously executes a file-based UDF associated with the specific email and ID.

    This function constructs a URL to run a UDF on a server, allowing for output data type customization for vector and raster outputs and supporting additional parameters for the UDF execution. If no ID is provided, the user's email is used as the identifier.

    Args:
        email (str): The user's email address, used to identify the user's saved UDFs. If the ID is not provided, this email will also be used as the ID.
        id (Optional[str]): Unique identifier for the UDF. If None, the function fetches the user's email as the ID.
        cache_max_age (Optional[int]): The maximum age when returning a result from the cache.
        _dtype_out_vector (str): Desired data type for vector output, defaults to a predefined type.
        _dtype_out_raster (str): Desired data type for raster output, defaults to a predefined type.
        _client_id (Optional[str]): Client identifier for API usage. If None, a default or global client ID may be used.
        **params: Additional keyword arguments for the UDF execution.

    Returns:
        A coroutine that, when awaited, sends a request to the server to execute the UDF and returns the server's response. The format and content of the response depend on the UDF's implementation and the server's response format.
    """
    if id is None:
        id = email
        email = context_get_user_email()

    url = f"{make_realtime_url(_client_id)}/api/v1/run/udf/saved/{email}/{id}"
    return await _run_file_async(
        url=url,
        cache_max_age=cache_max_age,
        _dtype_out_vector=_dtype_out_vector,
        _dtype_out_raster=_dtype_out_raster,
        **params,
    )


async def run_shared_file_async(
    token: str,
    *,
    cache_max_age: Optional[int] = None,
    _dtype_out_vector: str = DEFAULT_DTYPE_VECTOR,
    _dtype_out_raster: str = DEFAULT_DTYPE_RASTER,
    **params,
) -> Optional[Union["pd.DataFrame", "xr.Dataset"]]:
    """
    Asynchronously executes a shared file-based UDF using the specific access token.

    Constructs a URL to run an operation on a file accessible via a shared token, enabling customization of the output data types for vector and raster data. It accommodates additional parameters for the operation's execution.

    Args:
        token (str): A shared access token that authorizes the operation.
        cache_max_age (Optional[int]): The maximum age when returning a result from the cache.
        _dtype_out_vector (str): Desired data type for vector output, defaults to a predefined type.
        _dtype_out_raster (str): Desired data type for raster output, defaults to a predefined type.
        **params: Additional keyword arguments for the operation execution.

    Returns:
        A coroutine that, when awaited, sends a request to the server to execute the operation on the file and returns the server's response. The format and content of the response depend on the operation's implementation and the server's response format.
    """
    url = f"{make_shared_realtime_url(token)}/run/file"
    return await _run_file_async(
        url=url,
        cache_max_age=cache_max_age,
        _dtype_out_vector=_dtype_out_vector,
        _dtype_out_raster=_dtype_out_raster,
        _is_shared=True,
        **params,
    )



"""Vendored from xarray
https://github.com/pydata/xarray/blob/f4e0523be0ce8babaa8eff38365e5308b1fdb76b/xarray/core/formatting_html.py
"""

from __future__ import annotations

import json
import uuid
from functools import partial
from html import escape
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Sequence, Union

from pydantic import BaseModel
from xarray.core.formatting import short_data_repr
from xarray.core.options import _get_boolean_with_default

from fused_batch._formatter.common import copyable_text, icon, load_static_files
from fused_batch._formatter.udf import udf_section

if TYPE_CHECKING:
    from fused_batch.models.api import (
        JobConfig,
        JobStepConfig,
        JoinJobStepConfig,
        JoinSinglefileJobStepConfig,
        MapJobStepConfig,
        PartitionJobStepConfig,
        UdfJobStepConfig,
    )
    from fused_batch.models.internal import (
        AnyDatasetInput,
        AnyDatasetOutput,
        DatasetInput,
        DatasetInputV2,
        DatasetInputV2Table,
    )
    from fused_batch.models.udf import AnyBaseUdf


def short_data_repr_html(array):
    """Format "data" for DataArray and Variable."""
    internal_data = getattr(array, "variable", array)._data
    if hasattr(internal_data, "_repr_html_"):
        return internal_data._repr_html_()
    text = escape(short_data_repr(array))
    return f"<pre>{text}</pre>"


def format_dims(dims, dims_with_index):
    if not dims:
        return ""

    dim_css_map = {
        dim: " class='xr-has-index'" if dim in dims_with_index else "" for dim in dims
    }

    dims_li = "".join(
        f"<li><span{dim_css_map[dim]}>" f"{escape(str(dim))}</span>: {size}</li>"
        for dim, size in dims.items()
    )

    return f"<ul class='xr-dim-list'>{dims_li}</ul>"


def collapsible_section(
    name,
    inline_details: str = "",
    details: str = "",
    n_items: Optional[int] = None,
    enabled: bool = True,
    collapsed: bool = False,
):
    # "unique" id to expand/collapse the section
    data_id = "section-" + str(uuid.uuid4())

    has_items = n_items is not None and n_items
    n_items_span = "" if n_items is None else f" <span>({n_items})</span>"
    enabled_str = "" if enabled and has_items else "disabled"
    collapsed_str = "" if collapsed or not has_items else "checked"
    tip = " title='Expand/collapse section'" if enabled else ""

    return (
        f"<input id='{data_id}' class='xr-section-summary-in' "
        f"type='checkbox' {enabled_str} {collapsed_str}>"
        f"<label for='{data_id}' class='xr-section-summary' {tip}>"
        f"{name}:{n_items_span}</label>"
        f"<div class='xr-section-inline-details'>{inline_details}</div>"
        f"<div class='xr-section-details'>{details}</div>"
    )


def _mapping_section(
    mapping,
    name: str,
    details_func: Callable[[Any], str],
    max_items_collapse,
    expand_option_name,
    enabled: bool = True,
    **kwargs,
):
    n_items = len(mapping)
    expanded = _get_boolean_with_default(
        expand_option_name, n_items < max_items_collapse
    )
    collapsed = not expanded

    return collapsible_section(
        name,
        details=details_func(mapping, **kwargs),
        n_items=n_items,
        enabled=enabled,
        collapsed=collapsed,
    )


def _obj_repr(obj, header_components, sections, sections_class="xr-sections"):
    """Return HTML repr of an xarray object.

    If CSS is not injected (untrusted notebook), fallback to the plain text repr.

    """
    header = f"<div class='xr-header'>{''.join(h for h in header_components)}</div>"
    sections = "".join(
        f"<li class='xr-section-item'>{s}</li>" for s in sections if s is not None
    )

    icons_svg, css_style = load_static_files()
    return (
        "<div>"
        f"{icons_svg}<style>{css_style}</style>"
        f"<pre class='xr-text-repr-fallback'>{escape(repr(obj))}</pre>"
        "<div class='xr-wrap' style='display:none'>"
        f"{header}"
        f"<ul class='{sections_class}'>{sections}</ul>"
        "</div>"
        "</div>"
    )


def get_dataset_input_settings(input: AnyDatasetInput) -> str:
    results = ""
    for key, value in input.model_dump().items():
        if key == "tables":
            continue
        results += f"<li>{key}: <code>{value}</code></li>"
    return results


def _format_v2_table(table: DatasetInputV2Table) -> str:
    read_sidecar_files = " (w/ sidecar)" if table.read_sidecar_files else ""
    cache_locally = "(cache locally)" if table.cache_locally else ""
    return f"<code>{str(table.url)}</code>{read_sidecar_files}{cache_locally}"


def summarize_input(
    input_name: str,
    input: Union[str, DatasetInput],
    *,
    is_index: bool = False,
    dtype=None,
):
    is_dataset_input = not isinstance(input, str)
    cssclass_idx = " class='xr-has-index'" if is_index else ""
    dims_str = f"tables:{len(input.tables)}" if is_dataset_input else ""
    name = escape(str(input_name))
    dtype = ""

    # "unique" ids required to expand/collapse subsections
    attrs_id = "attrs-" + str(uuid.uuid4())

    # Convert DataFrame to HTML table
    # TODO: escape html characters # escape(inline_variable_array_repr(variable, 35))
    preview = f"[{', '.join(input.tables)}]" if is_dataset_input else ""

    attrs_ul = f"""
    <h4 style="margin-left: 20px;">Inputs</h4>
    <ul style="margin-top: 0; margin-bottom: 0px;">
    {''.join([f'<li><code>{t}</code></li>' for t in input.tables]) if is_dataset_input else f'<li>{input}</li>'}
    </ul>
    """
    if is_dataset_input:
        attrs_ul += f"""
            <h4 style="margin-left: 20px;">Settings</h4>
            <ul style="margin-top: 0; margin-bottom: 0px;">
            {get_dataset_input_settings(input)}
            </ul>
            """

    attrs_icon = icon("icon-file-text2")

    return (
        f"<div class='xr-var-name'><span{cssclass_idx}>{name}</span></div>"
        f"<div class='xr-var-dims'>{dims_str}</div>"
        f"<div class='xr-var-dtype'>{dtype}</div>"
        f"<div class='xr-var-preview xr-preview'>{preview}</div>"
        f"<input id='{attrs_id}' class='xr-var-attrs-in' "
        f"type='checkbox'>"
        f"<label for='{attrs_id}' title='Show/Hide attributes'>"
        f"{attrs_icon}</label>"
        f"<label></label>"
        f"<div class='xr-var-attrs'>{attrs_ul}</div>"
    )


def summarize_input_v2(
    input_name: str,
    input: DatasetInputV2,
    *,
    is_index: bool = False,
    dtype=None,
):
    cssclass_idx = " class='xr-has-index'" if is_index else ""
    dims_str = f"tables:{len(input.tables)}"
    name = escape(str(input_name))
    dtype = f"{input.operation.value}" if input.operation else "?"

    attrs_id = "attrs-" + str(uuid.uuid4())

    # TODO: escape html characters # escape(inline_variable_array_repr(variable, 35))
    preview = f"[{', '.join([t.url for t in input.tables])}]"

    attrs_ul = f"""
    <h4 style="margin-left: 20px;">Inputs</h4>
    <ul style="margin-top: 0; margin-bottom: 0px;">
    {''.join([f'<li>{_format_v2_table(t)}</li>' for t in input.tables])}
    </ul>
    """
    attrs_ul += f"""
        <h4 style="margin-left: 20px;">Settings</h4>
        <ul style="margin-top: 0; margin-bottom: 0px;">
        {get_dataset_input_settings(input)}
        </ul>
        """

    attrs_icon = icon("icon-file-text2")

    return (
        f"<div class='xr-var-name'><span{cssclass_idx}>{name}</span></div>"
        f"<div class='xr-var-dims'>{dims_str}</div>"
        f"<div class='xr-var-dtype'>{dtype}</div>"
        f"<div class='xr-var-preview xr-preview'>{preview}</div>"
        f"<input id='{attrs_id}' class='xr-var-attrs-in' "
        f"type='checkbox'>"
        f"<label for='{attrs_id}' title='Show/Hide attributes'>"
        f"{attrs_icon}</label>"
        f"<label></label>"
        f"<div class='xr-var-attrs'>{attrs_ul}</div>"
    )


def summarize_output(
    output_name: str,
    output: AnyDatasetOutput,
    *,
    is_index: bool = False,
    dtype=None,
):
    cssclass_idx = " class='xr-has-index'" if is_index else ""
    dims_str = ""
    name = escape(str(output_name))
    dtype = ""

    attrs_id = "attrs-" + str(uuid.uuid4())

    if hasattr(output, "url"):  # v2
        preview = f"{output.url}" if output.url else ""
    else:
        preview = repr(output)

    def get_dataset_output_settings(output: AnyDatasetOutput) -> str:
        results = ""
        for key, value in output.model_dump().items():
            results += f"<li>{key}: <code>{value}</code></li>"
        return results

    attrs_ul = f"""
        <h4 style="margin-left: 20px;">Settings</h4>
        <ul style="margin-top: 0; margin-bottom: 0px;">
        {get_dataset_output_settings(output)}
        </ul>
        """

    attrs_icon = icon("icon-file-text2")

    return (
        f"<div class='xr-var-name'><span{cssclass_idx}>{name}</span></div>"
        f"<div class='xr-var-dims'>{dims_str}</div>"
        f"<div class='xr-var-dtype'>{dtype}</div>"
        f"<div class='xr-var-preview xr-preview'>{preview}</div>"
        f"<input id='{attrs_id}' class='xr-var-attrs-in' "
        f"type='checkbox'>"
        f"<label for='{attrs_id}' title='Show/Hide attributes'>"
        f"{attrs_icon}</label>"
        f"<label></label>"
        f"<div class='xr-var-attrs'>{attrs_ul}</div>"
    )


def summarize_ingest_input(
    input: str,
    *,
    is_index: bool = False,
):
    cssclass_idx = " class='xr-has-index'" if is_index else ""
    name = escape(str(input))

    return (
        f'<div class="xr-var-name" style="grid-column: 1 / -2;"><span{cssclass_idx}>{name}</span></div>'
        f"{copyable_text(name, show_text=False)}"
    )


def summarize_inputs(
    inputs: List[Union[str, AnyDatasetInput]],
    names: Sequence[str] = ("input",),
):
    li_items = []
    if len(names) < len(inputs):
        names = ["input"] * len(inputs)

    for name, input in zip(names, inputs):
        if not isinstance(input, str) and input.type == "v2":  # DatasetInputType.V2
            # v2
            li_content = summarize_input_v2(name, input, is_index=False)
            li_items.append(f"<li class='xr-var-item'>{li_content}</li>")
        else:
            # v1 or string
            li_content = summarize_input(name, input, is_index=False)
            li_items.append(f"<li class='xr-var-item'>{li_content}</li>")

    vars_li = "".join(li_items)

    return f"<ul class='xr-var-list'>{vars_li}</ul>"


def summarize_ingest_inputs(inputs: List[str]):
    li_items = []
    for input in inputs:
        li_content = summarize_ingest_input(input, is_index=False)
        li_items.append(f"<li class='xr-var-item'>{li_content}</li>")

    vars_li = "".join(li_items)

    return f'<ul class="xr-var-list">{vars_li}</ul>'


def summarize_outputs(outputs: List[AnyDatasetOutput]):
    li_items = []
    for output in outputs:
        output_name = output.table if output.table else "(not set)"

        li_content = summarize_output(output_name, output, is_index=False)
        li_items.append(f"<li class='xr-var-item'>{li_content}</li>")

    vars_li = "".join(li_items)

    return f"<ul class='xr-var-list'>{vars_li}</ul>"


inputs_section = partial(
    _mapping_section,
    name="Inputs",
    details_func=summarize_inputs,
    max_items_collapse=25,
    expand_option_name="display_expand_coords",
)

ingest_inputs_section = partial(
    _mapping_section,
    name="Inputs",
    details_func=summarize_ingest_inputs,
    max_items_collapse=25,
    expand_option_name="display_expand_coords",
)

output_section = partial(
    _mapping_section,
    name="Outputs",
    details_func=summarize_outputs,
    max_items_collapse=25,
    expand_option_name="display_expand_coords",
)


def get_object_settings(obj: BaseModel, exclude: Sequence[str] = ()) -> str:
    def _format_kv(key: str, val: Any) -> str:
        if key in exclude:
            return ""
        return f"<li>{key}: <code>{val}</code></li>"

    return f"<ul>{''.join([_format_kv(key, val) for key, val in obj.model_dump().items()])}</ul>"


def _collect_udfs(
    config: JobStepConfig, *, names: Sequence[str] = ("udf",)
) -> Dict[str, AnyBaseUdf]:
    results = {}
    for udf_name in names:
        if hasattr(config, udf_name):
            udf_attr = getattr(config, udf_name)
            if udf_attr is not None:
                results[udf_attr.name] = udf_attr
    return results


def fused_ingestion_repr(ingest: PartitionJobStepConfig) -> str:
    obj_type = f"fused_batch.{ingest.__class__.__name__}: {ingest.name or ingest.output or ingest.output_metadata}"
    header_components = [
        f"<div class='xr-obj-type'>{escape(obj_type)}</div>",
        get_object_settings(ingest, exclude=["input", "type"]),
    ]

    inputs = ingest.input if isinstance(ingest.input, list) else [ingest.input]

    sections: List[str] = [
        ingest_inputs_section(inputs),
    ]

    return _obj_repr(ingest, header_components, sections)


def fused_udf_step_repr(step: UdfJobStepConfig) -> str:
    name = step.name
    if not name and step.udf is not None:
        name = step.udf.name
    obj_type = f"fused_batch.{step.__class__.__name__}: {name}"
    header_components = [
        f"<div class='xr-obj-type'>{escape(obj_type)}</div>",
        get_object_settings(step, exclude=["input", "udf", "type"]),
    ]

    inputs = [json.dumps(i) for i in step.input] if step.input is not None else []
    udfs = _collect_udfs(step)

    sections: List[str] = [
        ingest_inputs_section(inputs) if step.input is not None else None,
        udf_section(udfs),
    ]

    return _obj_repr(step, header_components, sections)


def fused_map_repr(map: MapJobStepConfig) -> str:
    output_name = map.output.url if hasattr(map.output, "url") else None
    obj_type = f"fused_batch.{map.__class__.__name__}: {map.name or output_name}"
    header_components = [
        f"<div class='xr-obj-type'>{escape(obj_type)}</div>",
        get_object_settings(map, exclude=["input", "output", "udf", "type"]),
    ]

    inputs = [map.input]
    outputs = [map.output]
    udfs = _collect_udfs(map)

    sections: List[str] = [
        inputs_section(inputs),
        output_section(outputs),
        udf_section(udfs),
    ]

    return _obj_repr(map, header_components, sections)


def fused_join_repr(join: JoinJobStepConfig) -> str:
    output_name = join.output.url if hasattr(join.output, "url") else None
    obj_type = f"fused_batch.{join.__class__.__name__}: {join.name or output_name}"
    header_components = [
        f"<div class='xr-obj-type'>{escape(obj_type)}</div>",
        get_object_settings(
            join,
            exclude=[
                "input_left",
                "input_right",
                "output",
                "udf",
                "type",
            ],
        ),
    ]

    inputs = [join.input_left, join.input_right]
    outputs = [join.output]
    udfs = _collect_udfs(join)

    sections: List[str] = [
        inputs_section(inputs, names=["left", "right"]),
        output_section(outputs),
        udf_section(udfs),
    ]

    return _obj_repr(join, header_components, sections)


def fused_join_singlefile_repr(join: JoinSinglefileJobStepConfig) -> str:
    output_name = join.output.url if hasattr(join.output, "url") else None
    obj_type = f"fused_batch.{join.__class__.__name__}: {join.name or output_name}"
    header_components = [
        f"<div class='xr-obj-type'>{escape(obj_type)}</div>",
        get_object_settings(
            join,
            exclude=[
                "input_left",
                "input_right",
                "output",
                "udf",
                "experimental_read_udf",
                "type",
            ],
        ),
    ]

    inputs = [join.input_left, join.input_right]
    outputs = [join.output]
    udfs = _collect_udfs(join, names=["experimental_read_udf", "udf"])

    sections: List[str] = [
        inputs_section(inputs, names=["left", "right"]),
        output_section(outputs),
        udf_section(udfs),
    ]

    return _obj_repr(join, header_components, sections)


def _get_job_repr(i: int, step: JobStepConfig) -> str:
    if hasattr(step, "_repr_html_"):
        html = step._repr_html_()
    else:
        html = repr(step)
    return f'<li><h4>Step {i}</h4><div style="padding-left: 2em;">{html}</div></li>'


def fused_job_repr(job: JobConfig) -> str:
    return f"""
    <div class='xr-obj-type'>fused.JobConfig: {job.name}</div>
    <ul>
    {''.join([_get_job_repr(step_idx, step) for step_idx, step in enumerate(job.steps)])}
    </ul>
    """



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Type, Union

if TYPE_CHECKING:
    from fused_batch.api import FusedAPI, FusedDockerAPI, FusedDockerHTTPAPI


API: Optional[Union[FusedAPI, FusedDockerAPI, FusedDockerHTTPAPI]] = None

# We expect this to be set after import
API_CLASS: Optional[Type[FusedAPI]] = None


def get_api(**api_kwargs) -> Union[FusedAPI, FusedDockerAPI, FusedDockerHTTPAPI]:
    global API
    if API is not None:
        return API

    if API_CLASS is not None:
        API = API_CLASS(**api_kwargs)
        return API

    raise ValueError("Internal error: Global API_CLASS was not set.")


def set_api(api: Union[FusedAPI, FusedDockerAPI, FusedDockerHTTPAPI]) -> None:
    global API
    API = api


def set_api_class(api_class: Type[FusedAPI]) -> None:
    global API_CLASS
    API_CLASS = api_class


def reset_api():
    global API
    API = None



from __future__ import annotations

from enum import Enum
from typing import Any, Dict, Literal, Optional, Sequence, Union

from pydantic import Field, PrivateAttr, RootModel
from typing_extensions import Annotated

from fused_batch.models.schema import Schema
from fused_batch.models.udf.base_udf import BaseUdf, UdfType

from .._inplace import _maybe_inplace


class JoinType(str, Enum):
    LEFT = "left"
    INNER = "inner"


class GeoPandasUdfV2(BaseUdf):
    """A user-defined function that operates on [`geopandas.GeoDataFrame`s][geopandas.GeoDataFrame]."""

    type: Literal[UdfType.GEOPANDAS_V2] = UdfType.GEOPANDAS_V2
    table_schema: Optional[Schema] = None
    """The [`Schema`][fused.models.Schema] describing the output of this UDF.
    """

    entrypoint: str
    """Name of the function within the code to invoke."""

    parameters: Dict[str, Any] = Field(default_factory=dict)
    """Parameters to pass into the entrypoint."""

    _parameter_list: Optional[Sequence[str]] = PrivateAttr(None)
    original_headers: Optional[str] = None

    _nested_callable = PrivateAttr(None)  # TODO : Find out type

    def set_parameters(
        self,
        parameters: Dict[str, Any],
        replace_parameters: bool = False,
        inplace: bool = False,
    ) -> "GeoPandasUdfV2":
        """Set the parameters on this UDF.

        Args:
            parameters: The new parameters dictionary.
            replace_parameters: If True, unset any parameters not in the parameters argument. Defaults to False.
            inplace: If True, modify this object. If False, return a new object. Defaults to True.
        """
        ret = _maybe_inplace(self, inplace)
        new_parameters = (
            parameters
            if replace_parameters
            else {
                **ret.parameters,
                **parameters,
            }
        )
        ret.parameters = new_parameters
        return ret

    def eval_schema(self, inplace: bool = False) -> "GeoPandasUdfV2":
        """Reload the schema saved in the code of the UDF.

        Note that this will evaluate the UDF function.

        Args:
            inplace: If True, update this UDF object. Otherwise return a new UDF object (default).
        """
        from fused_batch._udf.execute_v2 import execute_for_decorator

        new_udf = execute_for_decorator(self)
        assert isinstance(
            new_udf, GeoPandasUdfV2
        ), f"UDF has unexpected type: {type(new_udf)}"
        ret = _maybe_inplace(self, inplace)
        ret.table_schema = new_udf.table_schema
        ret._parameter_list = new_udf._parameter_list
        return ret

    def run_local(
        self,
        sample: Any | None = ...,
        *,
        inplace: bool = False,
        validate_output: bool = False,
        validate_imports: Optional[bool] = None,
        **kwargs,
    ) -> "UdfEvaluationResult":  # ruff: noqa: F821
        """Evaluate this UDF against a sample.

        Args:
            sample: Sample (from `get_sample`) to execute against.
            inplace: If True, update this UDF object with schema information. (default)
        """
        from fused_batch._udf.execute_v2 import execute_against_sample
        from fused_batch.models.udf._eval_result import (  # ruff: noqa: F401
            UdfEvaluationResult,
        )

        ret = _maybe_inplace(self, inplace)
        sample_list = [] if sample is Ellipsis else [sample]
        return execute_against_sample(
            udf=ret,
            input=sample_list,
            validate_output=validate_output,
            validate_imports=validate_imports,
            **kwargs,
        )


EMPTY_UDF = GeoPandasUdfV2(
    name="EMPTY_UDF", code="", entrypoint="", table_schema=Schema(fields=[])
)

AnyBaseUdf = Annotated[GeoPandasUdfV2, Field(..., discriminator="type")]


class RootAnyBaseUdf(RootModel[AnyBaseUdf]):
    pass



from __future__ import annotations

from typing import Optional, Tuple

from pydantic import ConfigDict

from fused.models.api.job import AnyJobStepConfig
from fused.models.base import FusedBaseModel
from fused.models.udf.udf import AnyBaseUdf

from ..request import WHITELISTED_INSTANCE_TYPES


class JobMetadata(FusedBaseModel):
    ec2_instance_type: Optional[WHITELISTED_INSTANCE_TYPES] = None
    """The EC2 instance this job is run on."""

    step_config: AnyJobStepConfig
    time_taken: Optional[float] = None
    """The time taken for the job, if known."""

    job_id: Optional[str] = None
    """The fused id for the job."""

    @property
    def job(self) -> AnyJobStepConfig:
        """The job step config that created this table."""
        return self.step_config

    @property
    def udf(self) -> Optional[AnyBaseUdf]:
        """The user-defined function that created this table."""
        if hasattr(self.step_config, "udf"):
            return self.step_config.udf

        return None

    @property
    def udf_code(self) -> Optional[str]:
        """The code string of the user-defined function that created this table."""
        udf = self.udf
        if udf is not None:
            return udf.code

        return None

    @property
    def inputs(self) -> Tuple:
        """The datasets that were combined to create this table."""
        if hasattr(self.step_config, "input"):
            return (self.step_config.input,)

    # We ignore extra keys because some keys are only useful for the backend
    model_config = ConfigDict(extra="ignore")



from typing import Mapping

from jinja2 import DictLoader


class RegisteredLoader(DictLoader):
    """Loader for templates specified by user code."""

    pass


class UrlLoader(DictLoader):
    """Loader for templates loaded by the user from a URL."""

    url: str

    def __init__(self, mapping: Mapping[str, str], url: str) -> None:
        self.url = url
        super().__init__(mapping)



from typing import TypeVar

import geopandas as gpd
import mercantile
import shapely

TileXYZ = TypeVar("TileXYZ", bound=mercantile.Tile)
TileGDF = TypeVar("TileGDF", bound=gpd.GeoDataFrame)
Bbox = TypeVar("Bbox", bound=shapely.geometry.polygon.Polygon)



import base64
from io import BytesIO
from typing import Iterable, Optional, Union

import geopandas as gpd
import geopandas.io.arrow
import pandas as pd
import pyarrow.parquet as pq
import shapely


def _normalize_url(url: str) -> str:
    if url.endswith("/"):
        return url[:-1]
    return url


def get_chunks_metadata(url: str) -> gpd.GeoDataFrame:
    """Returns a GeoDataFrame with each chunk in the table as a row.

    Args:
        url: URL of the table.
    """
    url = _normalize_url(url) + "/_sample"

    with pq.ParquetFile(url) as file:
        # do not use pq.read_metadata as it may segfault in versions >= 12 (tested on 15.0.1)
        sample_metadata = file.metadata

    if (
        b"fused:format_version" not in sample_metadata.metadata.keys()
        or sample_metadata.metadata[b"fused:format_version"] != b"5"
    ):
        raise ValueError(
            "Dataset does not have Fused metadata or it is an incompatible version."
        )

    metadata_bytes = sample_metadata.metadata[b"fused:_metadata"]
    metadata_bytes = base64.decodebytes(metadata_bytes)
    metadata_bio = BytesIO(metadata_bytes)
    chunks_df = pd.read_parquet(metadata_bio)
    return gpd.GeoDataFrame(
        chunks_df,
        geometry=chunks_df.apply(
            lambda row: shapely.box(
                xmin=row.bbox_minx,
                ymin=row.bbox_miny,
                xmax=row.bbox_maxx,
                ymax=row.bbox_maxy,
            ),
            axis=1,
        ),
    )


def get_chunk_from_table(
    url: str,
    file_id: Union[str, int, None],
    chunk_id: Optional[int],
    *,
    columns: Optional[Iterable[str]] = None,
) -> gpd.GeoDataFrame:
    """Returns a chunk from a table and chunk coordinates.

    This can be called with file_id and chunk_id from `get_chunks_metadata`.

    Args:
        url: URL of the table.
        file_id: File ID to read.
        chunk_id: Chunk ID to read.
    """
    if file_id is None:
        return gpd.read_parquet(url)
    else:
        url = _normalize_url(url) + f"/{file_id}.parquet"

    with pq.ParquetFile(url) as file:
        if chunk_id is not None:
            table = file.read_row_group(chunk_id, columns=columns)
        else:
            table = file.read(columns=columns)

        if b"geo" in table.schema.metadata:
            return geopandas.io.arrow._arrow_to_geopandas(table)
        else:
            return table.to_pandas()



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, TypeVar, Union
from urllib.parse import urlparse

if TYPE_CHECKING:
    from fused_batch.models.api.dataset import Table
    from fused_batch.models.internal.dataset import DatasetInputV2Table

T = TypeVar("T")


def is_url(path):
    prefixes = ["http://", "https://", "www."]
    return any(path.startswith(prefix) for prefix in prefixes)


def detect_passing_local_file_as_str(input: T) -> Union[T, Path]:
    if isinstance(input, str):
        try:
            parsed_url = urlparse(input)
            if parsed_url.scheme:
                # Do not do any rewriting if an explicit URL is passed in
                return input
        except:  # noqa: E722
            # Fall through to checking whether it exists as file
            pass

        as_path = Path(input)
        is_maybe_file = as_path.exists()

        if is_maybe_file:
            return as_path
    return input


def append_url_part(base: str, part: str) -> str:
    if base.endswith("/"):
        return f"{base[:-1]}/{part}"
    else:
        return f"{base}/{part}"


def table_to_name(table: Union[str, Table, DatasetInputV2Table]) -> str:
    # Table
    if hasattr(table, "name"):
        return table.name
    # DatasetInputV2Table
    if hasattr(table, "url"):
        table = table.url

    if table.endswith("/"):
        return table[:-1].rsplit("/", maxsplit=1)[-1]
    else:
        return table.rsplit("/", maxsplit=1)[-1]



"""Vendored from xarray
https://github.com/pydata/xarray/blob/f4e0523be0ce8babaa8eff38365e5308b1fdb76b/xarray/core/formatting_html.py
"""

from __future__ import annotations

from html import escape
from typing import TYPE_CHECKING, Dict, List, Tuple

from jinja2 import BaseLoader

from fused_batch._formatter.common import load_static_files
from fused_batch._formatter.template import template_section

if TYPE_CHECKING:
    from fused_batch._templates import FusedTemplate


def _obj_repr(obj, header_components, sections):
    """Return HTML repr of an xarray object.

    If CSS is not injected (untrusted notebook), fallback to the plain text repr.

    """
    header = f"<div class='xr-header'>{''.join(h for h in header_components)}</div>"
    sections = "".join(f"<li class='xr-section-item'>{s}</li>" for s in sections)

    icons_svg, css_style = load_static_files()
    return (
        "<div>"
        f"{icons_svg}<style>{css_style}</style>"
        f"<pre class='xr-text-repr-fallback'>{escape(repr(obj))}</pre>"
        "<div class='xr-wrap' style='display:none'>"
        f"{header}"
        f"<ul class='xr-sections'>{sections}</ul>"
        "</div>"
        "</div>"
    )


def fused_templates_repr(
    name: str, templates: Dict[str, Tuple[FusedTemplate, BaseLoader]]
) -> str:
    obj_type = f"fused_batch.templates.{name}"
    header_components = [f"<div class='xr-obj-type'>{escape(obj_type)}</div>"]

    sections: List[str] = [
        template_section(templates),
    ]
    return _obj_repr(templates, header_components, sections)


def fused_template_repr(template: FusedTemplate) -> str:
    obj_type = f"fused_batch.templates.{template.__class__.__name__}: {template.name}"
    header_components = [f"<div class='xr-obj-type'>{escape(obj_type)}</div>"]

    templates = {template.name: (template, None)}

    sections: List[str] = [
        template_section(templates),
    ]
    return _obj_repr(templates, header_components, sections)



# TODO: This file is no longer the most recent -- use fused.core.run_* instead
# This file is only for running non-saved (code included) UDFs

import base64
import json
import time
import warnings
from io import BytesIO
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, Dict, Optional, Union

import geopandas as gpd
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import requests
import xarray as xr
from loguru import logger
from PIL import Image

from fused_batch._environment import infer_display_method
from fused_batch._options import STAGING_DEFAULT_BASE_URL
from fused_batch._options import options as OPTIONS
from fused_batch._str_utils import detect_passing_local_file_as_str
from fused_batch._udf.execute_v2 import _transform_output
from fused_batch.api import FusedAPI
from fused_batch.models import AnyJobStepConfig
from fused_batch.models.udf._eval_result import UdfEvaluationResult
from fused_batch.models.udf.output import Output
from fused_batch.models.udf.udf import AnyBaseUdf
from fused_batch.warnings import FusedWarning

from ..core._impl._realtime_ops_impl import get_recursion_factor


def _resolve_udf_server_url(client_id: Optional[str] = None) -> str:
    if client_id and client_id.endswith("-staging"):
        return f"{STAGING_DEFAULT_BASE_URL}/realtime/{client_id}"

    if client_id is None:
        api = FusedAPI()
        client_id = api._automatic_realtime_client_id()

        if client_id is None:
            raise ValueError("Failed to detect realtime client ID")

    return f"{OPTIONS.base_url}/realtime/{client_id}"


def run_tile(
    x: float,
    y: float,
    z: float,
    data: Optional[
        Union[pd.DataFrame, gpd.GeoDataFrame, pa.Table, str, Path, Any]
    ] = None,
    step_config: Optional[AnyJobStepConfig] = None,
    params: Optional[Dict[str, str]] = None,
    *,
    print_time: bool = False,
    client_id: Optional[str] = None,
    dtype_out_vector: str = "parquet",
    dtype_out_raster: str = "tiff",
) -> UdfEvaluationResult:  # TODO: return png
    time_start = time.perf_counter()
    # We need to get the auth headers from the FusedAPI. Don't enable set_global_api
    # to avoid messing up the user's environment.
    api = FusedAPI(set_global_api=False)

    udf_server_url = _resolve_udf_server_url(client_id)

    assert step_config is not None
    # Apply parameters, if we want to step_config that's returned to have this,
    # overwrite step_config.
    step_config_with_params = step_config.set_udf(
        udf=step_config.udf, parameters=params
    )

    url = f"{udf_server_url}/api/v1/run/udf/tiles/{z}/{x}/{y}"

    # Headers
    recursion_factor = get_recursion_factor()
    headers = api._generate_headers({"Content-Type": "application/json"})
    headers["Fused-Recursion"] = f"{recursion_factor}"

    # Payload
    post_attr_json = {
        "data_left": data,
        "data_right": None,
        "step_config": step_config_with_params.model_dump_json(),
        "dtype_in": "json",
        "dtype_out_vector": dtype_out_vector,
        "dtype_out_raster": dtype_out_raster,
    }

    # Params
    req_params = {}

    # Make request
    start = time.time()

    r = requests.post(
        url=url,
        params=req_params,
        json=post_attr_json,
        headers=headers,
        timeout=OPTIONS.request_timeout,
    )

    end = time.time()
    if print_time:
        logger.info(f"Time in request: {end - start}")

    time_end = time.perf_counter()
    time_taken_seconds = time_end - time_start

    return _process_response(
        r, step_config=step_config_with_params, time_taken_seconds=time_taken_seconds
    )


def run(
    df_left: Optional[
        Union[pd.DataFrame, gpd.GeoDataFrame, pa.Table, str, Path, Any]
    ] = None,
    df_right: Optional[
        Union[pd.DataFrame, gpd.GeoDataFrame, pa.Table, str, Path]
    ] = None,
    step_config: Optional[AnyJobStepConfig] = None,
    params: Optional[Dict[str, str]] = None,
    *,
    print_time: bool = False,
    read_options: Optional[Dict] = None,
    client_id: Optional[str] = None,
    dtype_out_vector: str = "parquet",
    dtype_out_raster: str = "tiff",
) -> pd.DataFrame:
    """Run a UDF over a DataFrame.

    Args:
        df_left: Input DataFrame, or path to a local Parquet file.
        df_right: Input DataFrame, or path to a local Parquet file.
        step_config: AnyJobStepConfig.
        params: Additional parameters to pass to the UDF. Must be JSON serializable.

    Keyword Args:
        print_time: If True, print the amount of time taken in the request.
        read_options: If not None, options for reading `df` that will be passed to GeoPandas.
    """
    # TODO: This function is too complicated

    time_start = time.perf_counter()
    # We need to get the auth headers from the FusedAPI. Don't enable set_global_api
    # to avoid messing up the user's environment.
    api = FusedAPI(set_global_api=False)

    udf_server_url = _resolve_udf_server_url(client_id)

    assert step_config is not None
    # Apply parameters, if we want to step_config that's returned to have this,
    # overwrite step_config.
    step_config_with_params = step_config.set_udf(
        udf=step_config.udf, parameters=params
    )

    # Note: Custom UDF uses the json POST attribute.
    url = f"{udf_server_url}/api/v1/run/udf"

    # This is the body for when step_config_with_params.type == "udf".
    body = {
        "data_left": df_left,
        "step_config": step_config_with_params.model_dump_json(),
        "dtype_in": "json",
        "dtype_out_vector": dtype_out_vector,
        "dtype_out_raster": dtype_out_raster,
    }

    # TODO: This is not really supported
    if step_config_with_params.type != "udf":
        # Infer dtype_in from df_left. Ensure df_right, if present, is same type.
        if df_left is not None:
            data_left, dtype_in_left = _serialize_input(df_left, read_options)
            body["data_left"] = data_left
            body["dtype_in"] = dtype_in_left

        if df_right is not None:
            body["data_right"], dtype_in_right = _serialize_input(
                df_right, read_options
            )
            assert (
                dtype_in_left == dtype_in_right
            ), "Left and right must be same type. Fused currently supports 'GeoDataFrame' and 'geojson'."

    method = "POST"
    post_attr_json = body

    recursion_factor = get_recursion_factor()
    post_attr_headers = api._generate_headers({"Content-Type": "application/json"})
    post_attr_headers["Fused-Recursion"] = f"{recursion_factor}"

    req_params = {}

    # Make request
    start = time.time()

    r = requests.request(
        method=method,
        url=url,
        params=req_params,
        json=post_attr_json,
        headers=post_attr_headers,
        timeout=OPTIONS.request_timeout,
    )
    end = time.time()
    if print_time:
        logger.info(f"Time in request: {end - start}")

    time_end = time.perf_counter()
    time_taken_seconds = time_end - time_start

    return _process_response(
        r, step_config=step_config_with_params, time_taken_seconds=time_taken_seconds
    )


def _serialize_input(df: Union[str, pd.DataFrame, dict], read_options=None):
    if isinstance(df, pd.DataFrame):
        return (
            base64.encodebytes(_serialize_df(df, read_options).getvalue()).decode(
                "utf-8"
            ),
            "parquet",
        )
    elif isinstance(df, str):
        return json.loads(df), "geojson"
    elif isinstance(df, dict):
        return df, "geojson"


def _serialize_df(df: pd.DataFrame, read_options: Optional[Dict] = None):
    df = detect_passing_local_file_as_str(df)
    if isinstance(df, Path):
        read_options = {} if read_options is None else read_options
        if df.suffix == ".parquet":
            df = gpd.read_parquet(df, **read_options)
        else:
            df = gpd.read_file(df, **read_options)
    assert not isinstance(
        df, str
    ), f"A string input was passed but a file with that name could not be found: {df}"

    buf = BytesIO()
    if isinstance(df, pd.DataFrame):
        df.to_parquet(buf)
    elif isinstance(df, pa.Table):
        pq.write_table(df, buf)
    else:
        assert False, f"Unexpected type for df: {type(df)}"
    return buf


def _process_response(
    r: requests.Response,
    step_config: AnyJobStepConfig,
    time_taken_seconds: float,
) -> UdfEvaluationResult:
    result_content: Optional[bytes] = None
    output_df: Optional[pd.DataFrame] = None
    stdout: Optional[str] = None
    stderr: Optional[str] = None
    error_message: Optional[str] = None
    error_lineno: Optional[int] = None

    try:
        if r.status_code != 200 and r.status_code != 422:
            raise ValueError(r.text)
        result_content = r.content
        # x-fused-metadata holds LogHandler as JSON, which contains stdout/stderr.
        _fused_metadata = r.headers.get("x-fused-metadata")
        fused_metadata = json.loads(_fused_metadata) if _fused_metadata else {}

        # Extract stdout/stderr.
        stdout = fused_metadata.get("stdout")
        stderr = fused_metadata.get("stderr")

        # Extract udf.
        udf: Optional[AnyBaseUdf] = None
        if step_config is not None:
            udf = step_config.udf

        # Extract error line, if exists.
        error_lineno = fused_metadata.get("lineno")

        if r.status_code == 200:
            # If the UDF returned None.
            if len(result_content) == 0:
                return UdfEvaluationResult(
                    data=None,
                    udf=udf,
                    time_taken_seconds=time_taken_seconds,
                    stdout=stdout,
                    stderr=stderr,
                    error_message=error_message,
                    error_lineno=error_lineno,
                )

            # Else, process response output.
            res_buf = BytesIO(result_content)

            if r.headers["content-type"] == "image/png":
                display_method = infer_display_method(None, None)
                if display_method.show_widget:
                    from IPython.display import Image as IPythonImage
                    from IPython.display import display

                    display(IPythonImage(data=r.content, format="png"))

                image = Image.open(BytesIO(r.content))
                width, height = image.size
                if len(image.getbands()) == 1:
                    image_data = list(image.getdata())
                    image_data = [
                        image_data[i : i + width]
                        for i in range(0, len(image_data), width)
                    ]
                    data_array = xr.DataArray(image_data, dims=["y", "x"])
                else:
                    image_data = []
                    for band in range(len(image.getbands())):
                        band_data = list(image.getdata(band=band))
                        band_data = [
                            band_data[i : i + width]
                            for i in range(0, len(band_data), width)
                        ]
                        image_data.append(band_data)
                    data_array = xr.DataArray(image_data, dims=["band", "y", "x"])

                # Create the dataset with image, latitude, and longitude data
                dataset = xr.Dataset({"image": data_array})

                return UdfEvaluationResult(
                    data=dataset,
                    udf=udf,
                    time_taken_seconds=time_taken_seconds,
                    stdout=stdout,
                    stderr=stderr,
                    error_message=error_message,
                    error_lineno=error_lineno,
                )
            elif r.headers["content-type"] == "image/tiff":
                # TODO: Automatically display tiff?
                import rioxarray

                with NamedTemporaryFile(prefix="udf_result", suffix=".tiff") as ntmp:
                    with open(ntmp.name, "wb") as f:
                        f.write(r.content)
                    rda = rioxarray.open_rasterio(
                        ntmp.name,
                        masked=True,
                    )
                    dataset = xr.Dataset({"image": rda})

                    return UdfEvaluationResult(
                        data=dataset,
                        udf=udf,
                        time_taken_seconds=time_taken_seconds,
                        stdout=stdout,
                        stderr=stderr,
                        error_message=error_message,
                        error_lineno=error_lineno,
                    )
            else:  # assume parquet
                m = pq.read_metadata(res_buf)
                if b"geo" in m.metadata:
                    try:
                        output_df = gpd.read_parquet(res_buf)
                    except ValueError as e:
                        warnings.warn(
                            f"Result has geo metadata but could not be loaded in GeoPandas: {e}",
                            FusedWarning,
                        )
                if output_df is None:
                    output_df = pd.read_parquet(res_buf)

                if "fused_index" not in output_df.columns:
                    # TODO: Hack since the backend no longer responds with fused_index
                    output_df["fused_index"] = list(range(len(output_df)))

                new_output = _transform_output(
                    output=Output(data=output_df, skip_fused_index_validation=True)
                )

                new_output.validate_data_with_schema()

                if hasattr(udf, "table_schema") and udf.table_schema is None:
                    udf.table_schema = new_output.table_schema  # TODO: ?
                return UdfEvaluationResult(
                    data=new_output.data,
                    sidecar=new_output.sidecar_output,
                    udf=udf,
                    table_schema=new_output.table_schema,
                    time_taken_seconds=time_taken_seconds,
                    stdout=stdout,
                    stderr=stderr,
                    error_message=error_message,
                    error_lineno=error_lineno,
                )
        else:
            if "errormsg" in fused_metadata and fused_metadata["errormsg"]:
                error_message = f"The UDF returned the following error for chunk {fused_metadata['chunkinfo']} in line {fused_metadata.get('lineno')}:\n{fused_metadata['errormsg']}"
            elif "exception" in fused_metadata and fused_metadata["exception"]:
                error_message = fused_metadata["exception"]
            else:
                # No error message was returned, e.g. due to deserialization error
                try:
                    # Look for a "detail" field in the response payload
                    details_obj = json.loads(r.text)
                    error_message = details_obj["detail"]
                except:  # noqa: E722
                    error_message = r.text

            return UdfEvaluationResult(
                data=None,
                udf=udf,
                time_taken_seconds=time_taken_seconds,
                stdout=stdout,
                stderr=stderr,
                error_message=error_message,
                error_lineno=error_lineno,
            )
    except:  # noqa: E722
        r.raise_for_status()
        raise



"""Vendored from xarray
https://github.com/pydata/xarray/blob/f4e0523be0ce8babaa8eff38365e5308b1fdb76b/xarray/core/formatting_html.py
"""

from __future__ import annotations

import uuid
import warnings
from functools import partial
from html import escape
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional

import numpy as np
from xarray.core.formatting import short_data_repr
from xarray.core.options import _get_boolean_with_default

from fused_batch._formatter.common import (
    copyable_text,
    format_number,
    icon,
    load_static_files,
)
from fused_batch._formatter.noraise import noraise

if TYPE_CHECKING:
    from fused_batch.models.api import Dataset
    from fused_batch.models.api.dataset import Table


def short_data_repr_html(array):
    """Format "data" for DataArray and Variable."""
    internal_data = getattr(array, "variable", array)._data
    if hasattr(internal_data, "_repr_html_"):
        return internal_data._repr_html_()
    text = escape(short_data_repr(array))
    return f"<pre>{text}</pre>"


def format_dims(dims, dims_with_index):
    if not dims:
        return ""

    dim_css_map = {
        dim: " class='xr-has-index'" if dim in dims_with_index else "" for dim in dims
    }

    dims_li = "".join(
        f"<li><span{dim_css_map[dim]}>" f"{escape(str(dim))}</span>: {size}</li>"
        for dim, size in dims.items()
    )

    return f"<ul class='xr-dim-list'>{dims_li}</ul>"


def collapsible_section(
    name,
    inline_details: str = "",
    details: str = "",
    n_items: Optional[int] = None,
    enabled: bool = True,
    collapsed: bool = False,
):
    # "unique" id to expand/collapse the section
    data_id = "section-" + str(uuid.uuid4())

    has_items = n_items is not None and n_items
    n_items_span = "" if n_items is None else f" <span>({n_items})</span>"
    enabled_str = "" if enabled and has_items else "disabled"
    collapsed_str = "" if collapsed or not has_items else "checked"
    tip = " title='Expand/collapse section'" if enabled else ""

    return (
        f"<input id='{data_id}' class='xr-section-summary-in' "
        f"type='checkbox' {enabled_str} {collapsed_str}>"
        f"<label for='{data_id}' class='xr-section-summary' {tip}>"
        f"{name}:{n_items_span}</label>"
        f"<div class='xr-section-inline-details'>{inline_details}</div>"
        f"<div class='xr-section-details'>{details}</div>"
    )


def _mapping_section(
    mapping,
    name: str,
    details_func: Callable[[Any], str],
    max_items_collapse,
    expand_option_name,
    enabled: bool = True,
):
    n_items = len(mapping)
    expanded = _get_boolean_with_default(
        expand_option_name, n_items < max_items_collapse
    )
    collapsed = not expanded

    return collapsible_section(
        name,
        details=details_func(mapping),
        n_items=n_items,
        enabled=enabled,
        collapsed=collapsed,
    )


def _obj_repr(obj, header_components, sections):
    """Return HTML repr of an xarray object.

    If CSS is not injected (untrusted notebook), fallback to the plain text repr.

    """
    header = f"<div class='xr-header'>{''.join(h for h in header_components)}</div>"
    sections = "".join(f"<li class='xr-section-item'>{s}</li>" for s in sections)

    icons_svg, css_style = load_static_files()
    return (
        "<div>"
        f"{icons_svg}<style>{css_style}</style>"
        f"<pre class='xr-text-repr-fallback'>{escape(repr(obj))}</pre>"
        "<div class='xr-wrap' style='display:none'>"
        f"{header}"
        f"<ul class='xr-sections'>{sections}</ul>"
        "</div>"
        "</div>"
    )


def table_totals(table: Table):
    # TODO: maybe make this a method on the dataset?
    total_bounds = [np.inf, np.inf, -np.inf, -np.inf]

    sum_area = 0.0
    sum_length = 0.0
    sum_area_utm = 0.0
    sum_length_utm = 0.0
    sum_num_coords = 0
    sum_num_rows = 0

    if table.chunk_metadata:
        for meta in table.chunk_metadata:
            sum_area += meta.sum_area
            sum_length += meta.sum_length

            if meta.sum_area_utm is not None:
                sum_area_utm += meta.sum_area_utm

            if meta.sum_length_utm is not None:
                sum_length_utm += meta.sum_length_utm

            sum_num_coords += meta.num_coords
            sum_num_rows += meta.num_rows

            if meta.bbox_minx < total_bounds[0]:
                total_bounds[0] = meta.bbox_minx

            if meta.bbox_miny < total_bounds[1]:
                total_bounds[1] = meta.bbox_miny

            if meta.bbox_maxx > total_bounds[2]:
                total_bounds[2] = meta.bbox_maxx

            if meta.bbox_maxy > total_bounds[3]:
                total_bounds[3] = meta.bbox_maxy

    return (
        sum_area,
        sum_length,
        sum_area_utm,
        sum_length_utm,
        sum_num_coords,
        sum_num_rows,
        total_bounds,
    )


@noraise(incompat_version_message="Table formatting failed", default="")
def summarize_table(
    table_name: str,
    table: Optional[Table],
    *,
    is_index: bool = False,
    dtype=None,
):
    cssclass_idx = " class='xr-has-index'" if is_index else ""
    name = escape(str(table_name))
    if table is None:
        return (
            f"<div class='xr-var-name'><span{cssclass_idx}>{name}</span></div>"
            f"<div class='xr-var-dims'></div>"
            f"<div class='xr-var-dtype'></div>"
            f"<div class='xr-var-preview xr-preview'></div>"
            f"<div class='xr-var-attrs'></div>"
            f"<div class='xr-var-data'></div>"
        )
    dims_str = f"files:{format_number(table.num_files)}, chunks:{format_number(table.num_chunks)}"
    dtype = f"rows:{format_number(table.num_rows)}"

    # "unique" ids required to expand/collapse subsections
    attrs_id = "attrs-" + str(uuid.uuid4())
    data_id = "data-" + str(uuid.uuid4())
    chunk_repr_disabled = "disabled" if table.sample is None else ""

    # Convert DataFrame to HTML table
    # TODO: escape html characters # escape(inline_variable_array_repr(variable, 35))
    preview = f"[{', '.join(table.column_names)}]"

    def get_parent_inputs(table: Table):
        s = []

        if table.parent is None:
            return ""

        inputs = table.parent.inputs
        if len(inputs) == 1:
            if isinstance(inputs[0], list):
                if len(inputs[0]) < 6:
                    for input in inputs:
                        s.append(f"<li>Input: <code>{input}</code></li>")
                else:
                    n_more = len(inputs[0]) - 4
                    s.append(f"<li>Input: <code>{inputs[0][0]}</code></li>")
                    s.append(f"<li>Input: <code>{inputs[0][1]}</code></li>")
                    s.append(f"<li>... {n_more} more ...</li>")
                    s.append(f"<li>Input: <code>{inputs[0][-2]}</code></li>")
                    s.append(f"<li>Input: <code>{inputs[0][-1]}</code></li>")
            else:
                s.append(
                    f"<li>Input: <code>{dataset_to_url(table.parent.inputs[0])}</code></li>"
                )
        else:
            s.append(f"<li>Left Input: <code>{dataset_to_url(inputs[0])}</code></li>")
            s.append(f"<li>Right Input: <code>{dataset_to_url(inputs[1])}</code></li>")
        return "".join(s)
        # def dataset_to_name(inputs[0]):

    def dataset_to_url(dataset):
        if dataset is None:
            return None
        elif isinstance(dataset, str):
            return dataset
        elif dataset.type == "v2":  # DatasetInputType.V2
            return [str(t.url) for t in dataset.tables]
        else:
            return repr(dataset)

    if table.table_schema.fields:
        table_schema = [
            f"<li><code>{field.name}</code>: {field.type}</li>"
            for field in table.table_schema.fields
        ]
    else:
        # If we don't have the schema information, we can at least show the column names
        table_schema = [f"<li><code>{col}</code></li>" for col in table.column_names]

    def get_parent_copy(table: Table):
        if table.parent:
            copyable = copyable_text(
                f"fused_batch.experimental.job({repr(table.parent.job.model_dump_json())}, content_type='json')",
                show_text=False,
            )
            return f"<li>Copy job configuration as Python {copyable}</li>"
        else:
            return ""

    def parent_udf(table: Table):
        if table.parent and "udf" in table.parent.job.model_dump():
            return f"""<li>UDF name: <code>{table.parent.job.udf.name}</code></li>
                <li>Lines of code: {format_number(len(table.parent.job.udf.code.splitlines()))}</li>"""
        else:
            return "<li>There is no UDF.</li>"

    def get_parent_exec(table: Table):
        result = ""
        if table.parent and table.parent.time_taken:
            result += f"<li>Time taken (seconds): <code>{table.parent.time_taken:.2f}</code></li>"
        if table.parent and table.parent.job_id:
            result += f"<li>Job ID: {copyable_text(table.parent.job_id)}</li>"
        return result

    def get_parent_type(table: Table) -> str:
        if table.parent:
            return f"Parent: {table.parent.job.type}"
        else:
            return "Parent"

    attrs_ul = f"""
    <h4 style="margin-left: 20px;">Table Schema</h4>
    <ul style="margin-top: 0; margin-bottom: 0px;">
    {''.join(table_schema)}
    </ul>
    <h4 style="margin-left: 20px;">{get_parent_type(table)}</h4>
    <ul style="margin-top: 0; margin-bottom: 0px;">
    {get_parent_exec(table)}
    {get_parent_inputs(table)}
    {get_parent_copy(table)}
    </ul>
    <h4 style="margin-left: 20px;">UDF</h4>
    <ul style="margin-top: 0; margin-bottom: 0px;">
    {parent_udf(table)}
    </ul>
     """  # summarize_attrs(var.attrs)
    # {table.parent.udf_code}

    if hasattr(table, "sample") and table.sample is not None:
        sample = table.sample.copy()
        if "geometry" in sample.columns:
            with warnings.catch_warnings():
                # Ignore "Geometry column does not contain geometry."
                # Because we're converting the geometry column to a string
                warnings.filterwarnings(
                    action="ignore", message="Geometry column does not contain geometry"
                )
                sample["geometry"] = sample.geometry.map(lambda x: str(x)[:25] + "...")
        html_table = sample.to_html()
        data_repr = f'<div align="center" style="overflow-x:auto;">{html_table}</div>'
    else:
        data_repr = ""

    attrs_icon = icon("icon-file-text2")
    data_icon = icon("icon-database")

    return (
        f"<div class='xr-var-name'><span{cssclass_idx}>{name}</span></div>"
        f"<div class='xr-var-dims'>{dims_str}</div>"
        f"<div class='xr-var-dtype'>{dtype}</div>"
        f"<div class='xr-var-preview xr-preview'>{preview}</div>"
        f"<input id='{attrs_id}' class='xr-var-attrs-in' "
        f"type='checkbox'>"
        f"<label for='{attrs_id}' title='Show/Hide attributes'>"
        f"{attrs_icon}</label>"
        f"<input id='{data_id}' class='xr-var-data-in' type='checkbox' {chunk_repr_disabled}>"
        f"<label for='{data_id}' title='Show/Hide data repr'>"
        f"{data_icon}</label>"
        f"<div class='xr-var-attrs'>{attrs_ul}</div>"
        f"<div class='xr-var-data'>{data_repr}</div>"
    )


def summarize_tables(tables: Dict[str, Table]):
    li_items = []
    for table_name, table in tables.items():
        li_content = summarize_table(table_name, table, is_index=False)
        li_items.append(f"<li class='xr-var-item'>{li_content}</li>")

    vars_li = "".join(li_items)

    return f"<ul class='xr-var-list'>{vars_li}</ul>"


tables_section = partial(
    _mapping_section,
    name="Tables",
    details_func=summarize_tables,
    max_items_collapse=25,
    expand_option_name="display_expand_coords",
)


@noraise(incompat_version_message="Geometry formatting failed", default="")
def geometry_section(dataset: Dataset):
    geometry_table_name = "main" if "main" in dataset.tables else None
    if geometry_table_name is None and dataset.tables:
        geometry_table_name = sorted(dataset.tables.keys())[0]

    table = dataset.tables.get(geometry_table_name)
    return geometry_section_table(table)


@noraise(incompat_version_message="Geometry table formatting failed", default="")
def status_section_table(table: Table) -> str:
    if table.status:
        return collapsible_section(
            "Status",
            inline_details=f"""<span style="background-color: red;">{table.status}</span><br>Please check that the table exists.""",
            enabled=False,
            collapsed=True,
        )
    elif table.chunk_metadata is None:
        how_much_meta = "No" if table.num_files == 0 else "Minimal"
        return collapsible_section(
            "Status",
            inline_details=f"{how_much_meta} metadata loaded -- call <code>refresh()</code>",
            enabled=False,
            collapsed=True,
        )


@noraise(incompat_version_message="Geometry table formatting failed", default="")
def geometry_section_table(table: Optional[Table] = None):
    if table is not None:
        (
            sum_area,
            sum_length,
            sum_area_utm,
            sum_length_utm,
            sum_num_coords,
            sum_num_rows,
            total_bounds,
        ) = table_totals(table)

        return collapsible_section(
            "Geometry",
            inline_details=f"files:{format_number(table.num_files)}, chunks:{format_number(table.num_chunks)}, rows:{format_number(sum_num_rows)}, area: {sum_area_utm:.2E} m<sup>2</sup> length:{sum_length_utm:.2E} m <div>bbox=({total_bounds[0]:.6f}, {total_bounds[1]:.6f}, {total_bounds[2]:.6f}, {total_bounds[3]:.6f})</div>",
            enabled=False,
            collapsed=True,
        )
    else:
        return collapsible_section(
            "Geometry",
            inline_details="",
            enabled=False,
            collapsed=False,
        )


def fused_dataset_repr(dataset: Dataset) -> str:
    obj_type = f"fused_batch.Dataset: {dataset.base_path}"
    header_components = [f"<div class='xr-obj-type'>{escape(obj_type)}</div>"]

    sections: List[str] = [
        geometry_section(dataset),
        tables_section(dataset.tables),
    ]

    return _obj_repr(dataset, header_components, sections)


def fused_table_repr(table: Table) -> str:
    obj_type = f"fused_batch.Table: {table.name}"
    header_components = [f"<div class='xr-obj-type'>{escape(obj_type)}</div>"]

    sections: List[str] = [
        geometry_section_table(table),
        tables_section({table.name: table}),
    ]

    if table.status or table.chunk_metadata is None:
        sections.insert(0, status_section_table(table))

    return _obj_repr(table, header_components, sections)



from __future__ import annotations

import os
import warnings
from collections import defaultdict
from concurrent.futures import Future, ThreadPoolExecutor
from typing import Dict, List, Literal, Optional, Sequence, Set, Tuple, Union

import geopandas as gpd
import pandas as pd
import pyarrow as pa
from pydantic import (
    ConfigDict,
    Field,
    PrivateAttr,
    StrictStr,
    field_validator,
    model_validator,
)

from fused_batch._constants import DEFAULT_TABLE_NAMES
from fused_batch._formatter.formatter_dataset import (
    fused_dataset_repr,
    fused_table_repr,
)
from fused_batch._options import options as OPTIONS
from fused_batch.models._project_aware import FusedProjectAware
from fused_batch.models.api.enums import JoinType
from fused_batch.models.api.job import (
    AnyJobStepConfig,
    JoinJobStepConfig,
    JoinSinglefileJobStepConfig,
    MapJobStepConfig,
    _get_chunk,
    _get_chunk_v2,
    _get_file,
)
from fused_batch.models.base import FusedBaseModel
from fused_batch.models.internal.dataset import (
    DatasetInput,
    DatasetInputV2,
    DatasetOutput,
    DatasetOutputV2,
)
from fused_batch.models.schema import Schema
from fused_batch.models.udf.common import ChunkMetadata
from fused_batch.models.udf.udf import EMPTY_UDF, AnyBaseUdf
from fused_batch.models.urls import dataset_url_schema_validator
from fused_batch.warnings import FusedDefaultWarning, FusedRefreshingWarning

from ..request import WHITELISTED_INSTANCE_TYPES


class JobMetadata(FusedBaseModel):
    ec2_instance_type: Optional[WHITELISTED_INSTANCE_TYPES] = None
    """The EC2 instance this job is run on."""

    step_config: AnyJobStepConfig
    time_taken: Optional[float] = None
    """The time taken for the job, if known."""

    job_id: Optional[str] = None
    """The fused id for the job."""

    @property
    def job(self) -> AnyJobStepConfig:
        """The job step config that created this table."""
        return self.step_config

    @property
    def udf(self) -> Optional[AnyBaseUdf]:
        """The user-defined function that created this table."""
        if hasattr(self.step_config, "udf"):
            return self.step_config.udf

        return None

    @property
    def udf_code(self) -> Optional[str]:
        """The code string of the user-defined function that created this table."""
        udf = self.udf
        if udf is not None:
            return udf.code

        return None

    @property
    def inputs(self) -> Tuple[DatasetInput, ...]:
        """The datasets that were combined to create this table."""
        if hasattr(self.step_config, "input"):
            return (self.step_config.input,)

        if hasattr(self.step_config, "input_left") and hasattr(
            self.step_config, "input_right"
        ):
            return (self.step_config.input_left, self.step_config.input_right)

        return tuple()

    # We ignore extra keys because some keys are only useful for the backend
    model_config = ConfigDict(extra="ignore")


class Table(FusedBaseModel, FusedProjectAware):
    url: Optional[str] = None
    """URL of the table."""

    name: Optional[str] = None
    """The name of the table."""

    table_schema: Schema
    """The Schema representing this table."""

    parent: Optional[JobMetadata] = None
    """Metadata for the job that created this table."""

    column_names: List[StrictStr]
    """The list of column names in the table."""

    num_rows: Optional[int] = None
    """The number of rows in the table."""

    num_files: Optional[int] = None
    """The number of **non-empty** files."""

    num_chunks: Optional[int] = None
    """The number of **non-empty** chunk."""

    status: Optional[str] = None
    """A status of the table."""

    sample: Optional[pd.DataFrame] = Field(None, repr=False)

    chunk_metadata: Optional[List[ChunkMetadata]] = Field(..., repr=False)
    """Descriptive information about each chunk in this table"""

    _dataset: Optional["Dataset"] = PrivateAttr(None)

    @model_validator(mode="before")
    @classmethod
    def _default_name_from_url(cls, values):
        if "url" in values and "name" not in values:
            values["name"] = values["url"].rstrip("/").rsplit("/", maxsplit=1)[1]
        return values

    @property
    def columns(self) -> List[str]:
        """The list of columns in this table"""
        return self.column_names

    @property
    def metadata_gdf(self) -> gpd.GeoDataFrame:
        """The metadata of all chunks as a GeoDataFrame"""
        if self.chunk_metadata:
            return gpd.GeoDataFrame(
                [chunk.model_dump() for chunk in self.chunk_metadata],
                geometry=[chunk.to_box() for chunk in self.chunk_metadata],
            )
        else:
            raise ValueError("No chunk metadata loaded")

    def get_chunk(
        self, file_id: str | int | None = None, chunk_id: int | None = None
    ) -> gpd.GeoDataFrame:
        """Fetch a single chunk from this table

        Args:
            file_id: The identifier of this file. Defaults to None.
            chunk_id: The numeric index of the chunk within the file to fetch. Defaults to None.

        Returns:

            A `DataFrame` retrieved from the given file, chunk, and tables.
        """
        if self._dataset:
            return self._dataset.get_chunk(
                file_id=file_id, chunk_id=chunk_id, tables=[self.name]
            )
        elif self.url:
            return _get_chunk_v2(url=self.url, file_id=file_id, chunk_id=chunk_id)

        raise NotImplementedError("self._dataset and self.url are not set")

    def get_dataframe(
        self,
        file_id: str | int | None = None,
        chunk_id: int | None = None,
        parallel_fetch: bool = True,
    ) -> gpd.GeoDataFrame:
        """Fetch multiple chunks from this table as a dataframe

        Args:
            file_id: The identifier of the file to download. If `None` is passed, all files will be chosen. Defaults to 0.
            chunk_id: The numeric index of the chunk within the file to fetch. If `None` is passed, all chunks for the given file(s) will be chosen. Defaults to 0.
            parallel_fetch: Fetch in parallel. Defaults to True.

        Raises:
            ValueError: If the function would fetch more than `max_rows` rows.

        Returns:
            A `DataFrame` with all chunks concatenated together.
        """
        if self._dataset:
            return self._dataset.get_dataframe(
                file_id=file_id,
                chunk_id=chunk_id,
                tables=[self.name],
                parallel_fetch=parallel_fetch,
            )
        elif self.url:
            # TODO: Improve this and don't use a fake dataset. This will not work
            # if e.g. the table is at the root of a volume.
            dataset_base_path = self.url.rstrip("/").rsplit("/", maxsplit=1)[0]

            if self.chunk_metadata is None:
                warnings.warn(
                    "Refreshing project because chunk metadata was not found.",
                    FusedRefreshingWarning,
                )
                self.refresh()

            dataset = Dataset(base_path=dataset_base_path, tables={self.name: self})
            return dataset.get_dataframe(
                file_id=file_id,
                chunk_id=chunk_id,
                tables=[self.name],
                parallel_fetch=parallel_fetch,
            )

        raise NotImplementedError("self._dataset and self.url are not set")

    def get_dataframe_bbox(
        self,
        minx: float,
        miny: float,
        maxx: float,
        maxy: float,
        n_rows: Optional[int] = None,
        columns: Optional[List[str]] = None,
        clip: bool = True,
        buffer: Optional[float] = None,
    ) -> Union[pd.DataFrame, gpd.GeoDataFrame]:
        """Get a DataFrame of this Table of data in partitions matching the bounding box.

        Args:
            minx: Left coordinate of the box.
            miny: Bottom coordinate of the box.
            maxx: Right coordinate of the box.
            maxy: Top coordinate of the box.
            n_rows: If not None, up to this many rows will be returned.
            columns: If not None, only return these columns.
            clip: If True, only geometries that intersect the bounding box will be returned.
            buffer: If not None, this will be applied as the buffer for the partitions.
        """
        return self._api.download_table_bbox(
            path=self.url,
            minx=minx,
            miny=miny,
            maxx=maxx,
            maxy=maxy,
            n_rows=n_rows,
            columns=columns,
            clip=clip,
            buffer=buffer,
        )

    def join(
        self,
        other: Union[Table, str],
        output_table: Optional[str] = None,
        udf: Union[AnyBaseUdf, None] = None,
        *,
        how: Union[JoinType, Literal["left", "inner"]] = "inner",
        left_cache_locally: bool = False,
        right_cache_locally: bool = False,
        buffer_distance: Optional[float] = None,
    ) -> JoinJobStepConfig:
        """Construct a join config from two datasets

        Args:
            other: The other Dataset object to join on, or a path to another dataset object.
            output_table: Where to save the output of this operation. Defaults to `None`, which will not save the output.
            udf: The user-defined function to run in the join.

        Keyword Args:
            how: The manner of doing the join. Currently  Defaults to "inner".
            left_cache_locally: Whether to cache the left dataset locally in the join. Defaults to False.
            right_cache_locally: Whether to cache the right dataset locally in the join. Defaults to False.
            buffer_distance: The size of the buffer (in meters) on the left table to use during the join. Defaults to None.

        Examples:
            ```py
            import fused

            left_table = fused.open_table("s3://bucket/path/to/table")
            other_table = fused.open_table("s3://bucket/path/to/table")
            join_config = left_table.join(other_table)
            ```

        Returns:
            An object describing the join configuration.
        """
        parsed_udf = Dataset._parse_udf(udf)

        if isinstance(other, str):
            other = self._api.open_table(
                other,
                fetch_samples=False,
            )

        input_left = DatasetInputV2.from_table_url(
            self.url,
            cache_locally=left_cache_locally,
        )
        input_right = DatasetInputV2.from_table_url(
            other.url,
            cache_locally=right_cache_locally,
        )
        output = DatasetOutputV2(url=output_table)

        join_config = JoinJobStepConfig(
            udf=parsed_udf,
            input_left=input_left,
            input_right=input_right,
            output=output,
            buffer_distance=buffer_distance,
            how=how,
        )
        return join_config

    def join_singlefile(
        self,
        other: str,
        output_table: Optional[str] = None,
        udf: Union[AnyBaseUdf, None] = None,
        *,
        left_cache_locally: bool = False,
        buffer_distance: Optional[float] = None,
    ) -> JoinJobStepConfig:
        """Construct a join config from a dataset and a Parquet file URL.

        Args:
            other: The URL to the Parquet file to join all chunks with.
            output_table: Where to save the output of this operation. Defaults to `None`, which will not save the output.
            udf: The user-defined function to run in the join

        Keyword Args:
            left_cache_locally: Whether to cache the left dataset locally in the join. Defaults to False.

        Examples:
            ```py
            left_table = fused.open_table("s3://bucket/path/to/table")
            other_file = "s3://bucket/path/to/file.parquet"
            join_config = left_table.join_singlefile(other_file)
            ```

        Returns:
            An object describing the join configuration.
        """
        parsed_udf = Dataset._parse_udf(udf)

        input_left = DatasetInputV2.from_table_url(
            self.url,
            cache_locally=left_cache_locally,
        )
        output = DatasetOutputV2(url=output_table)

        join_config = JoinSinglefileJobStepConfig(
            udf=parsed_udf,
            input_left=input_left,
            input_right=other,
            output=output,
            buffer_distance=buffer_distance,
        )
        return join_config

    def map(
        self,
        output_table: Optional[str] = None,
        udf: Union[AnyBaseUdf, None] = None,
        *,
        cache_locally: bool = False,
    ) -> MapJobStepConfig:
        """Construct a `map` config from this Dataset

        Args:
            output_table: Where to save the output of this operation. Defaults to `None`, which will not save the output.
            udf: A user-defined function to run in this map. Defaults to None.

        Keyword Args:
            cache_locally: Advanced: whether to cache all the partitions locally in the map job. Defaults to False.

        Returns:
            An object describing the map configuration.
        """
        parsed_udf = Dataset._parse_udf(udf)

        dataset_input = DatasetInputV2.from_table_url(
            self.url,
            cache_locally=cache_locally,
        )
        output = DatasetOutputV2(url=output_table)

        map_config = MapJobStepConfig(
            udf=parsed_udf,
            input=dataset_input,
            output=output,
        )
        return map_config

    def _repr_html_(self) -> str:
        return fused_table_repr(self)

    def refresh(self, fetch_samples: Optional[bool] = None) -> Table:
        """Returns this table with updated metadata"""
        should_fetch_samples = False
        if fetch_samples is not None:
            should_fetch_samples = fetch_samples
        elif OPTIONS.open.fetch_samples is not None:
            should_fetch_samples = OPTIONS.open.fetch_samples

        new_table = self._api.open_table(
            path=self.url,
            fetch_samples=should_fetch_samples,
        )

        # Update self with all fields of new_table
        for key in new_table.model_dump().keys():
            value = getattr(new_table, key)
            setattr(self, key, value)

        return self

    model_config = ConfigDict(arbitrary_types_allowed=True)


class Dataset(FusedBaseModel, FusedProjectAware):
    """A class to describe everything that exists for a dataset in an environment"""

    base_path: StrictStr
    """The path on object storage where this dataset is stored"""

    tables: Dict[StrictStr, Optional[Table]] = Field(default_factory=dict)
    """The names of one or more attribute tables in this dataset"""

    def __getattribute__(self, key):
        try:
            return super().__getattribute__(key)
        except AttributeError:
            try:
                return self.tables[key]
            # Note that we need to raise an AttributeError, **not a KeyError** so that
            # IPython's _repr_html_ works here
            except KeyError:
                raise AttributeError(
                    f"object of type {type(self).__name__} has no attribute {key}"
                ) from None

    def __dir__(self) -> List[str]:
        """Provide method name lookup and completion. Only provide 'public'
        methods.
        """
        # This enables autocompletion
        extra_attrs = set(self.tables.keys())
        normal_dir = set(dir(type(self)))
        pydantic_fields = set(self.model_fields.keys())
        return sorted(normal_dir | pydantic_fields | extra_attrs)

    @field_validator("base_path")
    @classmethod
    def validate_base_path(cls, val):
        dataset_url_schema_validator.validate_python(val)
        return val

    @field_validator("tables")
    @classmethod
    def _validate_tables(cls, v):
        for name, table in v.items():
            if table is not None:
                table.name = name
        return v

    def _add_table_up_refs(self):
        for name, table in self.tables.items():
            # name is reset here just to be safe
            if table is not None:
                table.name = name
                table._dataset = self

    def join(
        self,
        other: Union[Dataset, str],
        output_table: Optional[str] = None,
        udf: Union[AnyBaseUdf, None] = None,
        *,
        how: Union[JoinType, Literal["left", "inner"]] = "inner",
        left_tables: Sequence[str] = DEFAULT_TABLE_NAMES,
        right_tables: Sequence[str] = DEFAULT_TABLE_NAMES,
        left_cache_locally: bool = False,
        right_cache_locally: bool = False,
        buffer_distance: Optional[float] = None,
    ) -> JoinJobStepConfig:
        """Construct a join config from two datasets

        Args:
            other: The other Dataset object to join on, or a path to another dataset object.
            output_table: Where to save the output of this operation. Defaults to `None`, which will not save the output.
            udf: The user-defined function to run in the join.

        Keyword Args:
            how: The manner of doing the join. Currently  Defaults to "inner".
            left_tables: The names of the attribute tables on the left side to include in the join. Defaults to ("main",).
            right_tables: The names of the attribute tables on the left side to include in the join. Defaults to ("main",).
            left_cache_locally: Whether to cache the left dataset locally in the join. Defaults to False.
            right_cache_locally: Whether to cache the right dataset locally in the join. Defaults to False.
            buffer_distance: The size of the buffer (in meters) on the left table to use during the join. Defaults to None.

        Examples:
            ```py
            import fused

            left_table = fused.open_table("s3://bucket/path/to/table")
            other_table = fused.open_table("s3://bucket/path/to/table")
            join_config = left_table.join(other_table)
            ```

        Returns:
            An object describing the join configuration.
        """
        parsed_udf = Dataset._parse_udf(udf)

        if isinstance(other, str):
            other = self._api.open_table(other, fetch_samples=False)

        self._validate_output(output_table)

        left_tables = [table for table in left_tables if table in self.tables]
        right_tables = [table for table in right_tables if table in other.tables]
        input_left = DatasetInput(
            base_path=self.base_path,
            tables=left_tables,
            cache_locally=left_cache_locally,
        )
        input_right = DatasetInput(
            base_path=other.base_path,
            tables=right_tables,
            cache_locally=right_cache_locally,
        )
        output = DatasetOutput(table=output_table)

        join_config = JoinJobStepConfig(
            udf=parsed_udf,
            input_left=input_left,
            input_right=input_right,
            output=output,
            buffer_distance=buffer_distance,
            how=how,
        )
        return join_config

    def join_singlefile(
        self,
        other: str,
        output_table: Optional[str] = None,
        udf: Union[AnyBaseUdf, None] = None,
        *,
        left_tables: Sequence[str] = DEFAULT_TABLE_NAMES,
        left_cache_locally: bool = False,
        buffer_distance: Optional[float] = None,
    ) -> JoinJobStepConfig:
        """Construct a join config from a dataset and a Parquet file URL.

        Args:
            other: The URL to the Parquet file to join all chunks with.
            output_table: Where to save the output of this operation. Defaults to `None`, which will not save the output.
            udf: The user-defined function to run in the join

        Keyword Args:
            left_tables: The names of the attribute tables on the left side to include in the join. Defaults to ("main",).
            left_cache_locally: Whether to cache the left dataset locally in the join. Defaults to False.

        Examples:
            ```py
            import fused

            left_table = fused.open_table("s3://bucket/path/to/table")
            other_file = "s3://bucket/path/to/file.parquet"
            join_config = left_table.join_singlefile(other_file)
            ```

        Returns:
            An object describing the join configuration.
        """
        parsed_udf = Dataset._parse_udf(udf)

        self._validate_output(output_table)

        left_tables = [table for table in left_tables if table in self.tables]
        input_left = DatasetInput(
            base_path=self.base_path,
            tables=left_tables,
            cache_locally=left_cache_locally,
        )
        output = DatasetOutput(table=output_table)

        join_config = JoinSinglefileJobStepConfig(
            udf=parsed_udf,
            input_left=input_left,
            input_right=other,
            output=output,
            buffer_distance=buffer_distance,
        )
        return join_config

    def map(
        self,
        output_table: Optional[str] = None,
        udf: Union[AnyBaseUdf, None] = None,
        *,
        tables: Sequence[str] = DEFAULT_TABLE_NAMES,
        cache_locally: bool = False,
    ) -> MapJobStepConfig:
        """Construct a `map` config from this Dataset

        Args:
            output_table: Where to save the output of this operation. Defaults to `None`, which will not save the output.
            udf: A user-defined function to run in this map. Defaults to None.

        Keyword Args:
            tables: The attribute tables to include in the map reduce. Defaults to ("main",).
            cache_locally: Advanced: whether to cache all the partitions locally in the map job. Defaults to False.

        Returns:
            An object describing the map configuration.
        """
        parsed_udf = Dataset._parse_udf(udf)

        self._validate_output(output_table)

        dataset_input = DatasetInput(
            base_path=self.base_path,
            tables=tables,
            cache_locally=cache_locally,
        )
        output = DatasetOutput(table=output_table)
        map_config = MapJobStepConfig(
            udf=parsed_udf,
            input=dataset_input,
            output=output,
        )
        return map_config

    def get_chunk(
        self,
        file_id: str | int | None = None,
        chunk_id: int | None = None,
        tables: Sequence[str] = DEFAULT_TABLE_NAMES,
    ) -> Union[pa.Table, pd.DataFrame, gpd.GeoDataFrame]:
        """Fetch a single chunk from this dataset

        Args:
            file_id: The identifier of this file. Defaults to None.
            chunk_id: The numeric index of the chunk within the file to fetch. Defaults to None.
            tables: the list of table names to fetch. Defaults to ("main",).

        Returns:

            A `DataFrame` retrieved from the given file, chunk, and tables.
        """
        return _get_chunk(
            base_path=self.base_path,
            tables=tables,
            file_id=file_id,
            chunk_id=chunk_id,
        )

    def get_dataframe(
        self,
        file_id: str | int | None = None,
        chunk_id: int | None = None,
        tables: Sequence[str] = DEFAULT_TABLE_NAMES,
        max_rows: int | None = 10_000_000,
        parallel_fetch: bool = True,
    ) -> pd.DataFrame:
        """Fetch multiple chunks from this dataset as a dataframe

        Args:
            file_id: The identifier of the file to download. If `None` is passed, all files will be chosen. Defaults to 0.
            chunk_id: The numeric index of the chunk within the file to fetch. If `None` is passed, all chunks for the given file(s) will be chosen. Defaults to 0.
            tables: the list of table names to fetch. Defaults to ("main",).
            max_rows: The maximum number of rows to fetch. If `None`, no limiting will be done. Defaults to 10_000_000.
            parallel_fetch: Fetch in parallel. Defaults to True.

        Raises:
            ValueError: If the function would fetch more than `max_rows` rows.

        Returns:
            A `DataFrame` with all chunks concatenated together.
        """
        # A mapping of file IDs to chunks that are available in this dataset
        first_table_name = tables[0]
        primary_table = self.tables[first_table_name]

        available_chunks: Dict[str, Set[int]] = defaultdict(set)
        if primary_table.chunk_metadata is not None:
            for meta in primary_table.chunk_metadata:
                available_chunks[meta.file_id].add(meta.chunk_id)
        else:
            warnings.warn(
                "Full metadata is not available, so defaulting to sequential file IDs.",
                FusedDefaultWarning,
            )
            for meta_id in range(primary_table.num_files):
                available_chunks[f"{meta_id}"] = {}

        # File ID to chunks that should be fetched
        # Note: this is not a defaultdict as I need to define the list of keys manually
        # so that I can loop over the `.keys()` below
        chunks_to_fetch: Dict[str, Set[int]] = {}

        # Initialize keys
        if file_id is not None:
            chunks_to_fetch[str(file_id)] = set()
        else:
            file_ids = available_chunks.keys()
            for file_id in file_ids:
                chunks_to_fetch[str(file_id)] = set()

        # A specific chunk is requested
        if chunk_id is not None:
            for file_id in chunks_to_fetch.keys():
                assert (
                    chunk_id in available_chunks[file_id]
                ), f"Chunk {chunk_id} does not exist for file {file_id}."
                chunks_to_fetch[file_id] = {chunk_id}

        else:
            # All chunks for the given file(s) are requested
            for file_id in chunks_to_fetch.keys():
                chunks_to_fetch[file_id] = available_chunks[file_id]

        # Check number of rows
        num_rows_to_fetch = 0
        if primary_table.chunk_metadata is not None:
            for meta in primary_table.chunk_metadata:
                file = chunks_to_fetch.get(meta.file_id, set())
                if meta.chunk_id in file:
                    num_rows_to_fetch += meta.num_rows
        else:
            num_rows_to_fetch = primary_table.num_rows

        if max_rows and num_rows_to_fetch > max_rows:
            raise ValueError(
                f"Cannot fetch {num_rows_to_fetch} given a maximum limit of {max_rows} rows. Pass a larger max_rows argument or specify fewer chunks."
            )

        chunk_dfs = []
        if parallel_fetch:
            with ThreadPoolExecutor(max_workers=OPTIONS.max_workers) as pool:
                futures: List[Future] = []
                if chunk_id is None:
                    for file_id in chunks_to_fetch.keys():
                        futures.append(
                            pool.submit(
                                _get_file,
                                base_path=self.base_path,
                                tables=tables,
                                file_id=file_id,
                            )
                        )
                    chunk_dfs = [future.result() for future in futures]
                else:
                    for file_id, chunk_ids in chunks_to_fetch.items():
                        for chunk_id in chunk_ids:
                            futures.append(
                                pool.submit(
                                    _get_chunk,
                                    base_path=self.base_path,
                                    tables=tables,
                                    file_id=file_id,
                                    chunk_id=chunk_id,
                                )
                            )
                    chunk_dfs = [future.result() for future in futures]
        else:
            if chunk_id is None:
                for file_id in chunks_to_fetch.keys():
                    file_df = _get_file(
                        base_path=self.base_path,
                        tables=tables,
                        file_id=file_id,
                    )
                    chunk_dfs.append(file_df)
            else:
                for file_id, chunk_ids in chunks_to_fetch.items():
                    for chunk_id in chunk_ids:
                        chunk_df = _get_chunk(
                            base_path=self.base_path,
                            tables=tables,
                            file_id=file_id,
                            chunk_id=chunk_id,
                        )
                        chunk_dfs.append(chunk_df)

        return pd.concat(chunk_dfs)

    def _repr_html_(self) -> str:
        return fused_dataset_repr(self)

    def _validate_output(self, output_table: Optional[str]):
        if output_table and output_table in self.tables:
            raise ValueError(
                f"Table {output_table} already exists. If you have removed it, you will need to refresh the dataset."
            )

    def delete(
        self, table_name: str, max_deletion_depth: int | Literal["unlimited"] = 2
    ) -> None:
        """Delete a table from the dataset.

        Specify a table to delete from the dataset using the provided table name.
        To prevent inadvertently deleting deeply nested objects in the object directory,
        users can specify the depth of deletion by specifying the maximum deletion depth.

        Args:
            table_name (str): The name of the table to be deleted from the dataset.
            max_deletion_depth (Optional[int], optional): The maximum depth of deletion.
                If provided, the deletion process will be performed only if there are no objects
                deeper than the specified depth. If set to "unlimited" deletion will be performed
                without any depth restrictions. Defaults to 2.

        Returns:
            None
        """
        # Construct path. Ends in `/` because it's a table.
        path = os.path.join(self.base_path, table_name, "")
        self._api.delete(path, max_deletion_depth)

    @staticmethod
    def _parse_udf(udf: Union[AnyBaseUdf, None]) -> AnyBaseUdf:
        if udf is None:
            return EMPTY_UDF
        elif isinstance(udf, AnyBaseUdf):
            return udf
        else:
            raise TypeError("Expected None or a Udf instance for `udf`.")

    model_config = ConfigDict(validate_assignment=True)



from io import BytesIO
from typing import Union

import geopandas as gpd
import pandas as pd


def parquet_to_df(b: bytes) -> Union[pd.DataFrame, gpd.GeoDataFrame]:
    try:
        return gpd.read_parquet(BytesIO(b))
    except:  # noqa: E722
        return pd.read_parquet(BytesIO(b))



from __future__ import annotations

from typing import Union
from uuid import UUID

from fused_batch.models import (
    JoinJobStepConfig,
    JoinSinglefileJobStepConfig,
    MapJobStepConfig,
    UdfJobStepConfig,
)
from fused_batch.models.api.dataset import JobMetadata, Table
from fused_batch.models.udf import AnyBaseUdf, RootAnyBaseUdf


def _fetch_udf_from_string(s: str) -> AnyBaseUdf:
    assert "gist.github.com" in s, f"'{s}' is not a GitHub Gist URL."
    gist_id = s.split("/")[-1]
    try:
        # Note that a github gist ID is an SHA hash, not a UUID, but they both have the
        # same number of bits as a hex string so this is easy validation.
        UUID(gist_id)
    except ValueError:
        raise ValueError(f"Could not parse '{s}' as a GitHub Gist URL")

    return AnyBaseUdf.from_gist(gist_id)


CoerceableToUdf = Union[
    AnyBaseUdf,
    dict,
    MapJobStepConfig,
    JoinJobStepConfig,
    JoinSinglefileJobStepConfig,
    UdfJobStepConfig,
    JobMetadata,
    Table,
]


def coerce_to_udf(udf: CoerceableToUdf) -> AnyBaseUdf:
    if isinstance(udf, AnyBaseUdf):
        return udf
    elif isinstance(udf, MapJobStepConfig):
        return udf.udf
    elif isinstance(udf, JoinJobStepConfig):
        return udf.udf
    elif isinstance(udf, JoinSinglefileJobStepConfig):
        return udf.udf
    elif isinstance(udf, UdfJobStepConfig):
        return udf.udf
    elif isinstance(udf, JobMetadata):
        return udf.udf
    elif isinstance(udf, Table):
        return udf.parent.udf
    else:
        return RootAnyBaseUdf.model_validate(udf).root



from __future__ import annotations

from http import HTTPStatus

from requests import HTTPError, Response


def raise_for_status(r: Response):
    """
    A wrapper around Response.raise_for_status to give more information on 422.
    """
    # Look for an error message from model parsing
    if r.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
        error_msg = r.json()
        raise HTTPError(error_msg, response=r)

    # Look for a Fused error message
    if 400 <= r.status_code < 600:
        to_raise: Exception | None = None
        try:
            if r.headers["Content-Type"] == "application/json":
                error_msg = r.json()
                if "message" in error_msg:
                    error_msg = error_msg["message"]
                to_raise = HTTPError(error_msg, response=r)
        except:  # noqa: E722
            # Fall through to raise_for_status, below
            pass
        if to_raise is not None:
            raise to_raise

    r.raise_for_status()



from typing import Literal, Optional

from fused._options import options as OPTIONS


def make_realtime_url(client_id: Optional[str]) -> str:
    from fused.api import FusedAPI

    if client_id is None:
        api = FusedAPI()
        client_id = api._automatic_realtime_client_id()

        if client_id is None:
            raise ValueError("Failed to detect realtime client ID")

    return f"{OPTIONS.base_url}/realtime/{client_id}"


def make_shared_realtime_url(id: str) -> str:
    return f"{OPTIONS.base_url}/realtime-shared/{id}"


def get_recursion_factor() -> int:
    return 1


def default_run_engine() -> Literal["remote", "local"]:
    from fused.api import AUTHORIZATION

    if OPTIONS.default_udf_run_engine is not None:
        return OPTIONS.default_udf_run_engine
    elif AUTHORIZATION.is_configured():
        return "remote"
    else:
        return "local"
